#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.dicom_tree_builder.py
"""Class and functions related to parsing a directory containing DICOM files and updating the GUI."""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
import os
import pydicom as dicom
from dicompylercore import dicomparser
from pydicom.errors import InvalidDicomError
from pubsub import pub
from threading import Thread
from queue import Queue
from dvha.db.dicom_parser import DICOM_Parser
from dvha.paths import ICONS
from dvha.tools.utilities import get_file_paths
from time import sleep


SCRIPT_DIR = os.path.dirname(__file__)


class DicomTreeBuilder:
    """
    This class processes data for various UI objects from models.import_dicom.ImportDicomFrame
    """
    def __init__(self, start_path, tree_ctrl_files, tree_ctrl_rois, tree_ctrl_roi_root, tree_ctrl_rois_images,
                 roi_map, search_subfolders=True):
        """
        :param start_path: directory to be scanned
        :type start_path: str
        :param tree_ctrl_files: tree in GUI used to visualize DICOM files
        :type tree_ctrl_files: CustomTreeCtrl
        :param tree_ctrl_rois: tree in GUI that lists rois of currently selected plan
        :type tree_ctrl_rois: CustomTreeCtrl
        :param tree_ctrl_roi_root: pointer to  the root node
        :param tree_ctrl_rois_images: a dictionary of image list pointers
        :type tree_ctrl_rois_images: dict
        :param roi_map: the object manager ROI mapping
        :type roi_map: tools.roi_name_manager.DatabaseROIs
        :param search_subfolders: indicates if files within sub-directories should be included
        :type search_subfolders: bool
        """

        # Store passed parameters
        self.start_path = start_path
        self.tree_ctrl_files = tree_ctrl_files
        self.tree_ctrl_rois = tree_ctrl_rois
        self.root_rois = tree_ctrl_roi_root
        self.tree_ctrl_rois_images = tree_ctrl_rois_images
        self.roi_map = roi_map
        self.search_subfolders = search_subfolders

        self.root_files = None
        self.tree_ctrl_files.DeleteAllItems()

        # These dictionaries will be used for easy lookup of tree_ctrl_files nodes
        self.patient_nodes = {}  # key: mrn
        self.study_nodes = {}  # key: study_instance_uid
        self.plan_nodes = {}  # key: sop_instance_uid
        self.file_nodes = {}  # key: absolute file path

        self.dicom_file_paths = {}
        self.other_dicom_files = {}
        self.current_index = 0
        self.file_types = ['rtplan', 'rtstruct', 'rtdose']
        self.file_tree = {}
        self.roi_name_map = {}
        self.roi_nodes = {}
        self.roi_name_over_ride = {}

        self.__do_subscribe()
        self.__set_images()
        self.__initialize_file_tree_root()

        self.parse_directory()

    def __do_subscribe(self):
        pub.subscribe(self.set_file_tree, "dicom_directory_parser_set_file_tree")

    def __set_images(self):
        self.image_list = wx.ImageList(16, 16)
        keys = ['rtplan', 'rtstruct', 'rtdose', 'other', 'studies', 'study', 'plan', 'patient']
        self.images = {key: self.image_list.Add(wx.Image(ICONS[key], wx.BITMAP_TYPE_PNG).Scale(16, 16).ConvertToBitmap())
                       for key in keys}

        self.tree_ctrl_files.AssignImageList(self.image_list)

    def __initialize_file_tree_root(self):
        self.root_files = self.tree_ctrl_files.AddRoot('Patients', ct_type=1)
        self.tree_ctrl_files.Expand(self.root_files)
        self.tree_ctrl_files.SetPyData(self.root_files, None)
        self.tree_ctrl_files.SetItemImage(self.root_files, self.images['studies'], wx.TreeItemIcon_Normal)

    def parse_directory(self):
        """
        Initiate directory parsing.  Creates a wx.Dialog, reads dicom headers and sorts files into dictionaries
        """
        PreImportParsingProgressFrame(self.start_path, search_subfolders=self.search_subfolders)

    def set_file_tree(self, tree, file_paths, other_dicom_files):
        """
        DicomDirectoryParserWorker subscribes to this function to initiate the tree_ctrl_files build
        :param tree: the plan_file_sets object from DicomDirectoryParserWorker
        :type tree: dict
        :param file_paths: the dicom_file_paths objet from DicomDirectoryParserWorker
        :type file_paths: dict
        """
        self.file_tree = tree
        self.dicom_file_paths = file_paths
        self.other_dicom_files = other_dicom_files
        wx.CallAfter(self.build_tree_ctrl_files)

    def get_file_type(self, dicom_dataset):
        """
        Get the file type of a DICOM dataset read by pydicom
        :param dicom_dataset: pydicom dataset
        :return: file_type among self.file_types or 'other'
        :rtype: str
        """
        file_type = dicom_dataset.Modality.lower()
        return [file_type, 'other'][file_type not in self.file_types]

    def build_tree_ctrl_files(self):
        """
        Build the tree for tree_ctrl_files and store their nodes using node adders.
        Then tell models.import_dicom.ImportDicomFrame to parse the data.
        """
        self.tree_ctrl_files.DeleteChildren(self.root_files)
        for key, patient in self.file_tree.items():
            self.add_patient_node(key)
            for study_uid, study in patient.items():
                self.add_study_node(key, study_uid)
                for plan_uid, plan in study.items():
                    self.add_plan_node(study_uid, plan_uid)
                    for file_type, file_list in self.dicom_file_paths[plan_uid].items():
                        if file_list:
                            for file_path in file_list:
                                if file_path not in self.file_nodes.keys():
                                    self.add_rt_file_node(plan_uid, file_type, file_path)
                                    break

        self.tree_ctrl_files.Expand(self.root_files)
        self.tree_ctrl_files.ExpandAllChildren(self.root_files)

        PreImportFileSetParserWorker(self.dicom_file_paths)

    def rebuild_tree_ctrl_rois(self, plan_uid):
        """
        Delete all nodes of current tree_ctrl_rois and build for the specified plan
        :param plan_uid: pydicom ds.SOPInstanceUID for the RT Plan of interest
        """
        self.tree_ctrl_rois.DeleteChildren(self.root_rois)
        if self.dicom_file_paths[plan_uid]['rtstruct'][0]:
            self.tree_ctrl_rois.SetItemBackgroundColour(self.root_rois, None)
            dicom_rt_struct = dicomparser.DicomParser(self.dicom_file_paths[plan_uid]['rtstruct'][0])
            structures = dicom_rt_struct.GetStructures()
            self.roi_name_map = {structures[key]['name']: {'key': key,
                                                           'type': structures[key]['type']}
                                 for key in list(structures) if structures[key]['type'] != 'MARKER'}
            self.roi_nodes = {}
            rois = list(self.roi_name_map)
            rois.sort()
            for roi in rois:
                self.roi_nodes[roi] = self.tree_ctrl_rois.AppendItem(self.root_rois, roi, ct_type=0)
                roi_type = [None, 'PTV'][self.roi_name_map[roi]['type'] == 'PTV']
                self.update_tree_ctrl_roi_with_roi_type(roi, roi_type=roi_type)
            self.apply_roi_name_over_rides(plan_uid)
        else:
            self.tree_ctrl_rois.SetItemBackgroundColour(self.root_rois, wx.Colour(255, 0, 0))

    def update_tree_ctrl_roi_with_roi_type(self, roi, roi_type=None):
        if roi_type is not None:
            text = '%s ----- ROI Type: %s' % (roi, roi_type)
        else:
            text = roi
        self.tree_ctrl_rois.SetItemText(self.roi_nodes[roi], text)

    def apply_roi_name_over_rides(self, plan_uid):
        if plan_uid in list(self.roi_name_over_ride):
            for roi_name, over_ride in self.roi_name_over_ride[plan_uid].items():
                if roi_name in list(self.roi_name_map):
                    self.roi_name_map[over_ride] = self.roi_name_map.pop(roi_name)
                if roi_name in list(self.roi_nodes):
                    self.roi_nodes[over_ride] = self.roi_nodes.pop(roi_name)
                    self.tree_ctrl_rois.SetItemText(self.roi_nodes[over_ride], over_ride)

    def set_roi_name_over_ride(self, plan_uid, roi_name, roi_name_over_ride):

        # get all plan_uids with this struct file (e.g., Eclipse plans with multiple Rxs)
        rt_file_path = self.dicom_file_paths[plan_uid]['rtstruct'][0]
        plan_uids = []
        for uid, file_paths in self.dicom_file_paths.items():
            if file_paths['rtstruct'] and file_paths['rtstruct'][0] == rt_file_path:
                plan_uids.append(uid)

        for uid in plan_uids:
            if uid not in list(self.roi_name_over_ride):
                self.roi_name_over_ride[uid] = {}
            self.roi_name_over_ride[uid][roi_name] = roi_name_over_ride

        self.apply_roi_name_over_rides(plan_uid)

    def add_patient_node(self, mrn):
        """
        Add a patient node to tree_ctrl_files
        :param mrn: pydicom's ds.PatientID
        :type mrn: str
        """
        if mrn not in list(self.patient_nodes):
            self.patient_nodes[mrn] = self.tree_ctrl_files.AppendItem(self.root_files, mrn, ct_type=1)
            # self.patient_nodes[mrn].Set3State(True)
            self.tree_ctrl_files.SetPyData(self.patient_nodes[mrn], None)
            self.tree_ctrl_files.SetItemImage(self.patient_nodes[mrn], self.images['patient'],
                                              wx.TreeItemIcon_Normal)

    def add_study_node(self, mrn, study_instance_uid):
        """
        Add a study node to tree_ctrl_files
        :param mrn: pydicom ds.PatientID
        :type mrn: str
        :param study_instance_uid: pydicom ds.StudyInstanceUID
        :type study_instance_uid: str
        """
        if study_instance_uid not in list(self.study_nodes):
            self.study_nodes[study_instance_uid] = self.tree_ctrl_files.AppendItem(self.patient_nodes[mrn],
                                                                                   study_instance_uid, ct_type=1)
            # self.study_nodes[uid].Set3State(True)
            self.tree_ctrl_files.SetPyData(self.study_nodes[study_instance_uid], None)
            self.tree_ctrl_files.SetItemImage(self.study_nodes[study_instance_uid], self.images['study'],
                                              wx.TreeItemIcon_Normal)

    def add_plan_node(self, study_instance_uid, plan_uid):
        """
        Add a plan node to tree_ctrl_files
        :param study_instance_uid: pydicom ds.StudyInstanceUID
        :type study_instance_uid: str
        :param plan_uid: pydicom ds.SOPInstanceUID of a RT Plan file
        :type plan_uid: str
        """
        if plan_uid not in list(self.plan_nodes):
            self.plan_nodes[plan_uid] = self.tree_ctrl_files.AppendItem(self.study_nodes[study_instance_uid], plan_uid, ct_type=1)
            # self.study_nodes[uid].Set3State(True)
            self.tree_ctrl_files.SetPyData(self.plan_nodes[plan_uid], None)
            self.tree_ctrl_files.SetItemImage(self.plan_nodes[plan_uid], self.images['plan'],
                                              wx.TreeItemIcon_Normal)

    def add_rt_file_node(self, plan_uid, file_type, file_path):
        """
        Add a file node to tree_ctrl_files
        :param plan_uid: pydicom ds.SOPInstanceUID of a RT Plan file
        :type plan_uid: str
        :param file_type: a type as specifed in self.file_types
        :type file_type: str
        :param file_path: the absolute file path of the DICOM file
        :type file_path: str
        """
        if file_path not in list(self.file_nodes):
            if file_type not in self.file_types:
                file_type = 'other'
            file_name = os.path.basename(file_path)
            self.file_nodes[file_path] = self.tree_ctrl_files.AppendItem(self.plan_nodes[plan_uid],
                                                                         "%s - %s" % (file_type, file_name), ct_type=0)
            # self.study_nodes[uid].Set3State(True)
            self.tree_ctrl_files.SetPyData(self.file_nodes[file_path], None)
            self.tree_ctrl_files.SetItemImage(self.file_nodes[file_path], self.images[file_type],
                                              wx.TreeItemIcon_Normal)

    def get_id_of_tree_ctrl_node(self, node):
        """
        :param node: item from of a wx.EVT_TREE_SEL_CHANGED on tree_ctrl_files
        :return: the node id and type if node found in the stored node dictionaries
        """
        for node_type in ['patient', 'study', 'plan', 'file']:
            for node_id, stored_node in getattr(self, "%s_nodes" % node_type).items():
                if node == stored_node:
                    return node_id, node_type
        return None, None

    def update_mapped_roi_status(self, physician, specific_roi=None):
        """
        Update the image next to the roi of the tree_ctrl_roi to indicate if the roi_name is found in the roi_map
        :param physician: name as stored in roi_map
        :type physician: str
        :param specific_roi: if specified, only this roi will be updated
        :type specific_roi: str
        """
        physician_is_valid = self.roi_map.is_physician(physician)
        if specific_roi is None:
            rois = self.roi_name_map.keys()
        else:
            rois = [specific_roi]  # wrap specific roi into a list
        for roi in rois:
            node = self.roi_nodes[roi]
            if physician_is_valid and self.roi_map.get_physician_roi(physician, roi) not in {'uncategorized'}:
                # self.tree_ctrl_rois.CheckItem(node, True)
                self.tree_ctrl_rois.SetItemImage(node, self.tree_ctrl_rois_images['yes'], wx.TreeItemIcon_Normal)
            else:
                # self.tree_ctrl_rois.CheckItem(node, False)
                self.tree_ctrl_rois.SetItemImage(node, self.tree_ctrl_rois_images['no'], wx.TreeItemIcon_Normal)

    def get_used_physician_rois(self, physician):
        """
        Get the physician rois that are used in the plan parsed by this object
        :param physician: name as stored in roi_map
        :type physician: str
        :return: used physician rois for this plan
        :rtype: list
        """
        if self.roi_map.is_physician(physician):
            return list(set([self.roi_map.get_physician_roi(physician, roi) for roi in self.roi_name_map.keys()]))
        return []

    @property
    def incomplete_plans(self):
        """
        Identify plans that do not include RT Structure and RT Dose files
        :return: the RT Plan SOP Instance UIDs of incomplete plans
        :rtype: list
        """
        incomplete_plan_uids = []
        for patient in self.file_tree.values():
            for study in patient.values():
                for plan_uid, plan in study.items():
                    files = [plan[file_type]['file_path'] for file_type in self.file_types]
                    if not all(files):
                        incomplete_plan_uids.append(plan_uid)
        return incomplete_plan_uids

    @property
    def checked_plans(self):
        """
        Identify the plans that are checked and return the abs_file_paths of their associated DICOM files
        :return: A dictionary with RT Plan SOP Instance UIDs for keys and dictionary of file_paths for values
        :rtype: dict
        """
        plans = {}
        for plan_node_uid, plan_node in self.plan_nodes.items():
            if self.tree_ctrl_files.IsItemChecked(plan_node):
                for mrn, study in self.file_tree.items():
                    for study_uid, plan in study.items():
                        for plan_uid, dicom_files in plan.items():
                            if plan_uid == plan_node_uid:
                                plans[plan_uid] = {file_type: file_obj['file_path']
                                                   for file_type, file_obj in dicom_files.items()}
        return plans


class PreImportParsingProgressFrame(wx.Dialog):
    """
    Create a window to display parsing progress and begin DicomDirectoryParserWorker. After thread completes,
    send message to begin DICOM parsing
    """
    def __init__(self, start_path, search_subfolders=True):
        """
        :param start_path: absolute path to begin parsing
        :type start_path: str
        :param search_subfolders: set to False to ignore sub-directories
        :type search_subfolders: bool
        """
        wx.Dialog.__init__(self, None)

        self.start_path = start_path
        self.search_subfolders = search_subfolders
        self.file_tree = {}

        self.gauge = wx.Gauge(self, wx.ID_ANY, 100)

        self.__set_properties()
        self.__do_subscribe()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.SetTitle("Reading DICOM Headers")
        self.SetMinSize((700, 100))

    def pub_set_title(self, msg):
        wx.CallAfter(self.SetTitle, msg)

    def __do_subscribe(self):
        pub.subscribe(self.update, "pre_import_progress_update")
        pub.subscribe(self.pub_set_title, "pre_import_progress_set_title")
        pub.subscribe(self.close, "pre_import_progress_close")

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_objects = wx.BoxSizer(wx.VERTICAL)
        self.label = wx.StaticText(self, wx.ID_ANY, "Progress Label:")
        sizer_objects.Add(self.label, 0, 0, 0)
        sizer_objects.Add(self.gauge, 0, wx.EXPAND, 0)
        sizer_wrapper.Add(sizer_objects, 0, wx.ALL | wx.EXPAND, 10)
        self.SetSizer(sizer_wrapper)
        self.Fit()
        self.Layout()
        self.Center()

    def update(self, msg):
        """
        Update the progress message and gauge
        :param msg: a dictionary with keys of 'label' and 'gauge' text and progress fraction, respectively
        :type msg: dict
        """
        wx.CallAfter(self.label.SetLabelText, msg['label'])
        wx.CallAfter(self.gauge.SetValue, int(100 * msg['gauge']))

    def run(self):
        """
        Initiate layout in GUI and begin dicom directory parser thread
        """
        self.Show()
        DicomDirectoryParserWorker(self.start_path, self.search_subfolders)

    def close(self):
        """
        Destroy layout in GUI and send message to being dicom parsing
        """
        pub.sendMessage("pre_import_complete")
        self.Destroy()


class DicomDirectoryParserWorker(Thread):
    """
    With a given start path, scan for RT DICOM files (plan, struct, dose) connected by SOPInstanceUID
    Previous versions strictly used StudyInstanceUID which was sufficient for Philips Pinnacle because
    it can export multiple prescriptions in a single RT Plan file, other TPS's export one file per plan
    """
    def __init__(self, start_path, search_subfolders):
        """
        :param start_path: initial directory path to scan for DICOM files
        :param search_subfolders: If true, files within all sub-directories will be included
        """
        Thread.__init__(self)
        self.start_path = start_path
        self.search_subfolders = search_subfolders
        self.file_types = ['rtplan', 'rtstruct', 'rtdose']
        self.req_tags = ['StudyInstanceUID', 'SOPInstanceUID', 'PatientName', 'PatientID']

        self.dicom_tag_values = {}
        self.dicom_files = {key: [] for key in self.file_types}
        self.plan_file_sets = {}
        self.uid_to_mrn = {}
        self.dicom_file_paths = {}
        self.other_dicom_files = {}

        self.start()  # begin thread

    def get_queue(self):
        file_paths = get_file_paths(self.start_path, search_subfolders=self.search_subfolders)
        file_count = len(file_paths)
        queue = Queue()

        for file_index, file_path in enumerate(file_paths):
            msg = {'label': "File Name: %s" % os.path.basename(file_path),
                   'gauge': file_index / file_count}
            queue.put((file_path, msg))

        return queue

    def run(self):
        """
        Begin the thread to parse directory. Returns plan_file_sets and dicom_file_paths through pubsub
        """

        queue = self.get_queue()
        worker = Thread(target=self.do_parse, args=[queue])
        worker.setDaemon(True)
        worker.start()
        queue.join()

        self.do_association()

        msg = {'label': "Complete",
               'gauge': 1.}
        wx.CallAfter(pub.sendMessage, "pre_import_progress_update", msg=msg)
        sleep(0.3)

        wx.CallAfter(pub.sendMessage, 'dicom_directory_parser_set_file_tree',
                     tree=self.plan_file_sets, file_paths=self.dicom_file_paths,
                     other_dicom_files=self.other_dicom_files)

    def do_parse(self, queue):
        while queue.qsize():
            parameters = queue.get()
            self.parser(*parameters)
            queue.task_done()

    def parser(self, file_path, msg):

        wx.CallAfter(pub.sendMessage, "pre_import_progress_update", msg=msg)

        file_name = os.path.basename(file_path)
        file_ext = os.path.splitext(file_path)[1]

        if file_ext and file_ext.lower() != '.dcm' or file_name.lower() == 'dicomdir':
            ds = None
        else:
            try:
                ds = dicom.read_file(file_path, stop_before_pixels=True, force=True)
            except InvalidDicomError:
                ds = None

        if ds is not None:

            if not self.is_data_set_valid(ds):
                print('Cannot parse %s\nOne of these tags is missing: %s' % (file_path, ', '.join(self.req_tags)))
            else:
                modality = ds.Modality.lower()
                timestamp = os.path.getmtime(file_path)

                self.dicom_tag_values[file_path] = {'timestamp': timestamp,
                                                    'study_instance_uid': ds.StudyInstanceUID,
                                                    'sop_instance_uid': ds.SOPInstanceUID,
                                                    'patient_name': ds.PatientName,
                                                    'mrn': ds.PatientID,
                                                    'modality': modality,
                                                    'matched': False}
                if modality not in self.file_types:
                    if ds.StudyInstanceUID not in self.other_dicom_files.keys():
                        self.other_dicom_files[ds.StudyInstanceUID] = []
                    self.other_dicom_files[ds.StudyInstanceUID].append(file_path)  # Store these to move after import
                else:
                    self.dicom_files[modality].append(file_path)

                    # All RT Plan files need to be found first
                    if modality == 'rtplan' and hasattr(ds, 'ReferencedStructureSetSequence'):
                        uid = ds.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID
                        mrn = self.dicom_tag_values[file_path]['mrn']
                        self.uid_to_mrn[uid] = ds.PatientID
                        self.dicom_tag_values[file_path]['ref_sop_instance'] = {'type': 'struct',
                                                                                'uid': uid}
                        study_uid = ds.StudyInstanceUID
                        plan_uid = ds.SOPInstanceUID
                        if mrn not in list(self.plan_file_sets):
                            self.plan_file_sets[mrn] = {}

                        if study_uid not in list(self.plan_file_sets[mrn]):
                            self.plan_file_sets[mrn][study_uid] = {}

                        self.plan_file_sets[mrn][study_uid][plan_uid] = {'rtplan': {'file_path': file_path,
                                                                                    'sop_instance_uid': plan_uid},
                                                                         'rtstruct': {'file_path': None,
                                                                                      'sop_instance_uid': None},
                                                                         'rtdose': {'file_path': None,
                                                                                    'sop_instance_uid': None}}
                        if plan_uid not in self.dicom_file_paths.keys():
                            self.dicom_file_paths[plan_uid] = {key: [] for key in self.file_types + ['other']}
                        self.dicom_file_paths[plan_uid]['rtplan'] = [file_path]

                    elif modality == 'rtdose':
                        uid = ds.ReferencedRTPlanSequence[0].ReferencedSOPInstanceUID
                        self.dicom_tag_values[file_path]['ref_sop_instance'] = {'type': 'plan',
                                                                                'uid': uid}
                    else:
                        self.dicom_tag_values[file_path]['ref_sop_instance'] = {'type': None,
                                                                                'uid': None}

    def do_association(self):
        # associate appropriate rtdose files to plans
        for file_index, dose_file in enumerate(self.dicom_files['rtdose']):
            dose_tag_values = self.dicom_tag_values[dose_file]
            ref_plan_uid = dose_tag_values['ref_sop_instance']['uid']
            study_uid = dose_tag_values['study_instance_uid']
            mrn = dose_tag_values['mrn']
            for plan_file_set in self.plan_file_sets[mrn][study_uid].values():
                plan_uid = plan_file_set['rtplan']['sop_instance_uid']
                if plan_uid == ref_plan_uid:
                    self.dicom_tag_values[dose_file]['matched'] = True
                    plan_file_set['rtdose'] = {'file_path': dose_file,
                                               'sop_instance_uid': dose_tag_values['sop_instance_uid']}
                    if 'rtdose' in self.dicom_file_paths[plan_uid].keys():
                        self.dicom_file_paths[plan_uid]['rtdose'].append(dose_file)
                    else:
                        self.dicom_file_paths[plan_uid]['rtdose'] = [dose_file]

        # associate appropriate rtstruct files to plans
        for mrn_index, mrn in enumerate(list(self.plan_file_sets)):
            for study_uid in list(self.plan_file_sets[mrn]):
                for plan_uid, plan_file_set in self.plan_file_sets[mrn][study_uid].items():
                    plan_file = plan_file_set['rtplan']['file_path']
                    ref_struct_uid = self.dicom_tag_values[plan_file]['ref_sop_instance']['uid']
                    for struct_file in self.dicom_files['rtstruct']:
                        struct_uid = self.dicom_tag_values[struct_file]['sop_instance_uid']
                        if struct_uid == ref_struct_uid:
                            self.dicom_tag_values[plan_file]['matched'] = True
                            plan_file_set['rtstruct'] = {'file_path': struct_file,
                                                         'sop_instance_uid': struct_uid}
                            if 'rtstruct' in self.dicom_file_paths[plan_uid].keys():
                                self.dicom_file_paths[plan_uid]['rtstruct'].append(struct_file)
                            else:
                                self.dicom_file_paths[plan_uid]['rtstruct'] = [struct_file]

        # find unmatched structure and dose files, pair by StudyInstanceUID to plan if plan doesn't have struct/dose
        for dcm_file, tags in self.dicom_tag_values.items():
            modality = tags['modality']
            if not tags['matched'] and modality in {'rtstruct', 'rtdose'}:
                for mrn_index, mrn in enumerate(list(self.plan_file_sets)):
                    for study_uid in list(self.plan_file_sets[mrn]):
                        for plan_uid, plan_file_set in self.plan_file_sets[mrn][study_uid].items():
                            if study_uid == tags['study_instance_uid']:
                                if modality not in plan_file_set.keys():
                                    self.dicom_file_paths[plan_uid][modality] = [dcm_file]
                                else:
                                    self.dicom_file_paths[plan_uid][modality].append(dcm_file)

    def is_data_set_valid(self, ds):
        for tag in self.req_tags:
            if not hasattr(ds, tag):
                return False
        return True


class PreImportFileSetParserWorker(Thread):
    def __init__(self, file_paths):
        Thread.__init__(self)

        pub.sendMessage("pre_import_progress_set_title", msg='Parsing File Sets')

        self.file_paths = file_paths
        self.total_plan_count = len(self.file_paths)

        self.start()

    def run(self):
        queue = self.get_queue()
        worker = Thread(target=self.do_parse, args=[queue])
        worker.setDaemon(True)
        worker.start()
        queue.join()
        sleep(0.3)  # Allow time for user to see final progress in GUI
        pub.sendMessage('pre_import_progress_close')

    def get_queue(self):
        queue = Queue()
        for plan_counter, uid in enumerate(list(self.file_paths)):
            if self.file_paths[uid]['rtplan'] and self.file_paths[uid]['rtstruct'] and self.file_paths[uid]['rtdose']:
                init_params = {'plan_file': self.file_paths[uid]['rtplan'][0],
                               'structure_file': self.file_paths[uid]['rtstruct'][0],
                               'dose_file': self.file_paths[uid]['rtdose'][0]}
                msg = {'label': 'Parsing File Set %s of %s' % (plan_counter+1, self.total_plan_count),
                       'gauge': plan_counter / self.total_plan_count}
                queue.put((uid, init_params, msg))
        return queue

    def do_parse(self, queue):
        while queue.qsize():
            parameters = queue.get()
            self.parser(*parameters)
            queue.task_done()

        plan_count = self.total_plan_count
        msg = {'label': 'Parsing Complete: %s fileset%s' % (plan_count, ['', 's'][plan_count != 1]),
               'gauge': 1.}
        pub.sendMessage("pre_import_progress_update", msg=msg)

    def parser(self, uid, init_params, msg):
        pub.sendMessage("pre_import_progress_update", msg=msg)

        pre_import_data = DICOM_Parser(**init_params).pre_import_data
        msg = {'uid': uid, 'init_params': pre_import_data}
        pub.sendMessage("set_pre_import_parsed_dicom_data", msg=msg)



import warnings
with warnings.catch_warnings():  # ignore Install python-Levenshtein warning
    warnings.simplefilter("ignore")
    from fuzzywuzzy import fuzz

from dvha.tools.roi_name_manager import clean_name


class ROINamePredictor:
    def __init__(self, roi_map, weight_simple=1., weight_partial=0.6, threshold=0):
        """
        :param roi_map: ROI map object
        :type roi_map: DatabaseROIs
        """
        self.roi_map = roi_map
        norm_weight = weight_partial + weight_simple
        self.weight = {'simple': 2 * weight_simple / norm_weight,
                       'partial': 2 * weight_partial / norm_weight}
        self.threshold = threshold

    def get_best_roi_match(self, roi, physician, return_score=False):
        physician_variations = self.roi_map.get_all_variations_of_physician(physician)
        fuzz_scores = self.get_combined_fuzz_scores(roi, physician_variations)
        if fuzz_scores:
            predicted_variation, score = fuzz_scores[0][1], fuzz_scores[0][0]
            prediction = self.roi_map.get_physician_roi(physician, predicted_variation)
            if score > self.threshold:
                if return_score:
                    return prediction, score
                return prediction

    def get_combined_fuzz_score(self, a, b, mode='geom_mean'):
        a, b = clean_name(a), clean_name(b)

        simple = float(fuzz.ratio(a, b) * self.weight['simple'])
        partial = float(fuzz.partial_ratio(a, b) * self.weight['partial'])

        return self.combine_scores(simple, partial, mode=mode)

    @staticmethod
    def combine_scores(score_1, score_2, mode='average'):
        if mode == 'geom_mean':
            return (score_1 * score_2) ** 0.5
        elif mode == 'product':
            return score_1 * score_2 / 100.
        else:  # average
            return (score_1 + score_2) / 2.

    def get_combined_fuzz_scores(self, string, list_of_strings):
        scores = [self.get_combined_fuzz_score(string, string_b) for string_b in list_of_strings]
        if scores:
            order_index = sorted(range(len(scores)), key=lambda k: scores[k])
            return [(scores[i], list_of_strings[i]) for i in order_index[::-1]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.machine_learning.py
"""
Classes to view and calculate Machine Learning predictions
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from functools import partial
import numpy as np
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor
from sklearn.model_selection import train_test_split
from sklearn.svm import SVR
from sklearn.tree import DecisionTreeRegressor
from dvha.dialogs.export import save_data_to_file
from dvha.options import DefaultOptions
from dvha.paths import MODELS_DIR
from dvha.models.plot import PlotMachineLearning, PlotFeatureImportance
from dvha.tools.errors import ErrorDialog
from dvha.tools.stats import MultiVariableRegression
from dvha.tools.utilities import set_msw_background_color, get_window_size, load_object_from_file, set_frame_icon


class MachineLearningFrame(wx.Frame):
    def __init__(self, main_app_frame, data, title, regressor=None, tool_tips=None, include_test_data=True):
        wx.Frame.__init__(self, None)

        self.main_app_frame = main_app_frame
        self.data = data
        self.title = title
        self.regressor = [regressor, ALGORITHMS[title]['regressor']][regressor is None]
        self.tool_tips = [tool_tips, ALGORITHMS[title]['tool_tips']][tool_tips is None]
        self.include_test_data = include_test_data

        self.reg = None
        self.plot = PlotMachineLearning(self, ml_type=self.title, ml_type_short=self.ml_type_short,
                                        include_test_data=include_test_data, **self.data)

        self.feature_importance_dlg = None

        self.input = {}
        self.defaults = {}
        self.getters = {}

        self.data_split_input = {'test_size': wx.TextCtrl(self, wx.ID_ANY, "0.25"),
                                 'train_size': wx.TextCtrl(self, wx.ID_ANY, "None"),
                                 'random_state': wx.TextCtrl(self, wx.ID_ANY, "None"),
                                 'shuffle': wx.ComboBox(self, wx.ID_ANY, choices=["True", "False"],
                                                        style=wx.CB_DROPDOWN | wx.CB_READONLY)}

        self.data_split_defaults = {'test_size': 0.25,
                                    'train_size': None,
                                    'random_state': None,
                                    'shuffle': True}

        self.data_split_getters = {'test_size': self.to_float_or_none,
                                   'train_size': self.to_float_or_none,
                                   'random_state': self.to_int_or_none,
                                   'shuffle': self.to_bool}

        self.button_calculate = wx.Button(self, wx.ID_ANY, "Calculate")
        self.button_importance = wx.Button(self, wx.ID_ANY, "Importance Plot")
        self.button_export_data = wx.Button(self, wx.ID_ANY, "Export Data")
        self.button_save_html = wx.Button(self, wx.ID_ANY, "Save Plot to HTML")
        self.button_save_svg = wx.Button(self, wx.ID_ANY, "Save Plot to SVG")
        self.button_save_model = wx.Button(self, wx.ID_ANY, "Save Model")

        self.do_bind()

    def do_bind(self):
        self.Bind(wx.EVT_BUTTON, self.on_calculate, id=self.button_calculate.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_feature_importance, id=self.button_importance.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_export, id=self.button_export_data.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_save_html, id=self.button_save_html.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_save_svg, id=self.button_save_svg.GetId())
        self.Bind(wx.EVT_BUTTON, self.on_save_model, id=self.button_save_model.GetId())
        self.Bind(wx.EVT_SIZE, self.on_resize)

    def set_properties(self):
        self.SetTitle(self.title)
        x_size = [0.6, 0.8][self.include_test_data]
        self.SetMinSize(get_window_size(x_size, 0.7))
        self.set_defaults()

        for key, input_obj in self.input.items():
            input_obj.SetToolTip(self.tool_tips[key])

        for key, input_obj in self.data_split_input.items():
            input_obj.SetToolTip(DATA_SPLIT_TOOL_TIPS[key])

    def do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.HORIZONTAL)
        sizer_side_bar = wx.BoxSizer(wx.VERTICAL)
        sizer_actions = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Actions"), wx.VERTICAL)
        sizer_param = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Parameters"), wx.VERTICAL)
        sizer_split_param = wx.StaticBoxSizer(wx.StaticBox(self, wx.ID_ANY, "Data Split"), wx.VERTICAL)

        variables = list(self.input)
        variables.sort()
        sizer_input = {variable: wx.BoxSizer(wx.HORIZONTAL) for variable in variables}
        for variable in variables:
            sizer_input[variable].Add(wx.StaticText(self, wx.ID_ANY, "%s:\t" % variable), 0, wx.EXPAND, 0)
            sizer_input[variable].Add(self.input[variable], 1, wx.EXPAND, 0)
            sizer_param.Add(sizer_input[variable], 1, wx.EXPAND | wx.ALL, 2)
        sizer_side_bar.Add(sizer_param, 0, wx.ALL | wx.EXPAND, 5)

        split_variables = ['test_size', 'train_size', 'random_state', 'shuffle']
        sizer_split_input = {variable: wx.BoxSizer(wx.HORIZONTAL) for variable in split_variables}
        for variable in split_variables:
            sizer_split_input[variable].Add(wx.StaticText(self, wx.ID_ANY, "%s:\t" % variable), 0, wx.EXPAND, 0)
            sizer_split_input[variable].Add(self.data_split_input[variable], 1, wx.EXPAND, 0)
            sizer_split_param.Add(sizer_split_input[variable], 1, wx.EXPAND | wx.ALL, 2)
        sizer_side_bar.Add(sizer_split_param, 0, wx.ALL | wx.EXPAND, 5)

        sizer_actions.Add(self.button_calculate, 1, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5)
        sizer_actions.Add(self.button_importance, 1, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5)
        sizer_actions.Add(self.button_export_data, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_actions.Add(self.button_save_html, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_actions.Add(self.button_save_svg, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_actions.Add(self.button_save_model, 1, wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT, 5)
        sizer_side_bar.Add(sizer_actions, 0, wx.ALL | wx.EXPAND, 5)

        sizer_wrapper.Add(sizer_side_bar, 0, wx.EXPAND, 0)

        sizer_wrapper.Add(self.plot.layout, 1, wx.EXPAND, 0)

        set_msw_background_color(self)  # If windows, change the background color
        set_frame_icon(self)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def get_param(self, variable):
        ans = variable in list(self.input)
        input_dict = [self.data_split_input, self.input][ans]
        default_dict = [self.data_split_defaults, self.defaults][ans]
        getters_dict = [self.data_split_getters, self.getters][ans]

        try:
            return getters_dict[variable](input_dict[variable].GetValue())
        except Exception:
            return default_dict[variable]

    def set_defaults(self):
        for variable, input_obj in self.input.items():
            input_obj.SetValue(self.to_str_for_gui(self.defaults[variable]))
        for variable, input_obj in self.data_split_input.items():
            input_obj.SetValue(self.to_str_for_gui(self.data_split_defaults[variable]))

    @property
    def input_parameters(self):
        return {variable: self.get_param(variable) for variable in self.input.keys()
                if self.get_param(variable) != self.defaults[variable]}

    def set_input_parameters(self, input_parameters):
        for variable, value in input_parameters.items():
            self.input[variable].SetValue(str(value))

    @property
    def data_split_parameters(self):
        return {variable: self.get_param(variable) for variable in self.data_split_input.keys()
                if self.get_param(variable) != self.data_split_defaults[variable]}

    def set_data_split_parameters(self, data_split_input):
        for variable, value in data_split_input.keys():
            self.data_split_input[variable].SetValue(str(value))

    def to_int_float_or_none(self, str_value):
        if str_value.lower() == 'none':
            return None
        return self.to_int_or_float(str_value)

    @staticmethod
    def to_int_float_string_or_none(str_value):
        if str_value.lower() == 'none':
            return None
        if str_value.isnumeric():  # int
            return int(float(str_value))
        if '.' in str_value and len(str_value.split('.') == 2):
            return float(str_value)
        return str_value

    @staticmethod
    def to_int_or_float(str_value):
        if str_value.isnumeric():  # int
            return int(float(str_value))
        return float(str_value)

    def to_int_or_none(self, str_value):
        if str_value.lower() == 'none':
            return None
        return self.to_int(str_value)

    def to_float_or_none(self, str_value):
        if str_value.lower() == 'none':
            return None
        return self.to_float(str_value)

    @staticmethod
    def to_int(str_value):
        return int(float(str_value))

    @staticmethod
    def to_float(str_value):
        return float(str_value)

    @staticmethod
    def to_str(str_value):
        return str_value

    @staticmethod
    def to_float_or_str(str_value):
        if '.' in str_value and len(str_value.split('.') == 2):
            if ("%s" % (str_value.split('.')[0] + str_value.split('.')[1])).isnumeric():
                return float(str_value)
        return str_value

    def to_float_str_or_none(self, str_value):
        try:
            return self.to_float_or_none(str_value)
        except Exception:
            return str_value

    @staticmethod
    def to_bool(str_value):
        if str_value.lower() == 'true':
            return True
        return False

    @staticmethod
    def to_bool_or_str(str_value):
        if str_value.lower() == 'true':
            return True
        if str_value.lower() == 'false':
            return False
        return str_value

    @staticmethod
    def to_str_for_gui(value):
        if value is None:
            return 'None'
        return str(value)

    @property
    def plot_data(self):
        try:
            self.reg = self.regressor(**self.input_parameters)
            return MachineLearningPlotData(self.data['X'], self.data['y'], self.reg, **self.data_split_parameters)
        except Exception as e:
            wx.MessageBox(str(e), 'Error!',
                          wx.OK | wx.OK_DEFAULT | wx.ICON_WARNING)

    def do_regression(self):
        self.reg = self.regressor(**self.input_parameters)
        return MachineLearningPlotData(self.data['X'], self.data['y'], self.reg, **self.data_split_parameters)

    def on_calculate(self, evt):
        data = self.plot_data
        if data is not None:
            self.plot.update_data(data)

    def redraw_plot(self):
        self.plot.redraw_plot()

    def on_resize(self, *evt):
        try:
            self.Refresh()
            self.Layout()
            wx.CallAfter(self.redraw_plot)
        except RuntimeError:
            pass

    def on_export(self, evt):
        save_data_to_file(self, 'Save machine learning data to csv', self.plot.get_csv())

    def on_save_html(self, *evt):
        title = 'Save %s Plot to .html' % self.title.title()
        try:
            if self.main_app_frame.export_figure is None:
                save_data_to_file(self, title, self.plot.html_str,
                                  initial_dir="",wildcard="HTML files (*.html)|*.html")
            else:
                self.save_fig_with_attr('html')
        except Exception as e:
            ErrorDialog(self, str(e), "Save Error")

    def on_save_svg(self, *evt):
        title = 'Save %s Plot to .svg' % self.title.title()
        try:
            if self.main_app_frame.export_figure is None:
                save_data_to_file(self, title, self.plot.export_svg,
                                  initial_dir="", data_type='function', wildcard="SVG files (*.svg)|*.svg")
            else:
                self.save_fig_with_attr('svg')
        except Exception as e:
            ErrorDialog(self, str(e), "Save Error")

    def save_fig_with_attr(self, format_):
        title = 'Save %s Plot to .%s' % (self.title.title(), format_)
        fig_attr_dict = self.main_app_frame.export_figure.fig_attr_dict
        func = partial(self.plot.save_figure, format_, fig_attr_dict)
        save_data_to_file(self, title, func, initial_dir="", data_type='function',
                          wildcard="%s files (*.%s)|*.%s" % (format_.upper(), format_, format_))

    def on_save_model(self, evt):
        data = {'y_variable': self.plot.y_variable,
                'regression': self.reg,
                'regressor': self.regressor,
                'tool_tips': self.tool_tips,
                'x_variables': self.plot.x_variables,
                'title': self.title,
                'input_parameters': self.input_parameters,
                'data_split': self.data_split_parameters,
                'version': DefaultOptions().VERSION}
        save_data_to_file(self, 'Save Model', data,
                          wildcard="MODEL files (*.mlr)|*.mlr", data_type='pickle', initial_dir=MODELS_DIR)

    def run(self):
        self.set_properties()
        self.do_layout()
        self.Show()
        self.on_calculate(None)

    @property
    def frame_size(self):
        return self.GetSize()

    @property
    def ml_type_short(self):
        return ''.join([s[0] for s in self.title.split(' ')]).upper()

    def on_feature_importance(self, evt):
        title = "Importance Figure for %s (%s)" % (self.title, self.data['y_variable'])
        plot_title = "%s Feature Importances for %s" % (self.title, self.data['y_variable'])
        self.feature_importance_dlg = FeatureImportanceFrame(self.data['options'], self.data['x_variables'],
                                                             self.reg.feature_importances_, title, plot_title)
        self.feature_importance_dlg.Show()


class RandomForestFrame(MachineLearningFrame):
    def __init__(self, main_app_frame, data, include_test_data=True):
        MachineLearningFrame.__init__(self, main_app_frame, data, 'Random Forest', include_test_data=include_test_data)

        self.input = {'n_estimators': wx.TextCtrl(self, wx.ID_ANY, "100"),
                      'criterion': wx.ComboBox(self, wx.ID_ANY, choices=["mse", "mae"],
                                               style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'max_depth': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'min_samples_split': wx.TextCtrl(self, wx.ID_ANY, "2"),
                      'min_samples_leaf': wx.TextCtrl(self, wx.ID_ANY, "1"),
                      'min_weight_fraction_leaf': wx.TextCtrl(self, wx.ID_ANY, "0"),
                      'max_features': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'max_leaf_nodes': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'min_impurity_decrease': wx.TextCtrl(self, wx.ID_ANY, "0"),
                      'bootstrap': wx.ComboBox(self, wx.ID_ANY, choices=["True", "False"],
                                               style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'oob_score': wx.ComboBox(self, wx.ID_ANY, choices=["True", "False"],
                                               style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'n_jobs': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'random_state': wx.TextCtrl(self, wx.ID_ANY, "None")}

        self.defaults = {'n_estimators': 100,
                         'criterion': 'mse',
                         'max_depth': None,
                         'min_samples_split': 2,
                         'min_samples_leaf': 1,
                         'min_weight_fraction_leaf': 0.,
                         'max_features': None,
                         'max_leaf_nodes': None,
                         'min_impurity_decrease': 0.,
                         'bootstrap': True,
                         'oob_score': False,
                         'n_jobs': None,
                         'random_state': None}

        self.getters = {'n_estimators': self.to_int,
                        'criterion': self.to_str,
                        'max_depth': self.to_int_or_none,
                        'min_samples_split': self.to_int_or_float,
                        'min_samples_leaf': self.to_int_or_float,
                        'min_weight_fraction_leaf': self.to_float,
                        'max_features': self.to_int_float_string_or_none,
                        'max_leaf_nodes': self.to_int_or_none,
                        'min_impurity_decrease': self.to_float,
                        'bootstrap': self.to_bool,
                        'oob_score': self.to_bool,
                        'n_jobs': self.to_int_or_none,
                        'random_state': self.to_int_or_none}

        self.run()


class GradientBoostingFrame(MachineLearningFrame):
    def __init__(self, main_app_frame, data, include_test_data=True):
        MachineLearningFrame.__init__(self, main_app_frame, data, 'Gradient Boosting',
                                      include_test_data=include_test_data)

        self.input = {'loss': wx.ComboBox(self, wx.ID_ANY, choices=["ls", "lad", "huber", "quantile"],
                                          style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'learning_rate': wx.TextCtrl(self, wx.ID_ANY, "0.1"),
                      'n_estimators': wx.TextCtrl(self, wx.ID_ANY, "100"),
                      'subsample': wx.TextCtrl(self, wx.ID_ANY, "1.0"),
                      'criterion': wx.ComboBox(self, wx.ID_ANY, choices=["friedman_mse", "mse", "mae"],
                                               style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'max_depth': wx.TextCtrl(self, wx.ID_ANY, "3"),
                      'min_samples_split': wx.TextCtrl(self, wx.ID_ANY, "2"),
                      'min_samples_leaf': wx.TextCtrl(self, wx.ID_ANY, "1"),
                      'min_weight_fraction_leaf': wx.TextCtrl(self, wx.ID_ANY, "0"),
                      'max_features': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'alpha': wx.TextCtrl(self, wx.ID_ANY, "0.9"),
                      'max_leaf_nodes': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'min_impurity_decrease': wx.TextCtrl(self, wx.ID_ANY, "0"),
                      'init': wx.ComboBox(self, wx.ID_ANY, choices=["DummyEstimator", "zero"],
                                          style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'random_state': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'presort': wx.ComboBox(self, wx.ID_ANY, choices=["auto", "True", "False"],
                                             style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'validation_fraction': wx.TextCtrl(self, wx.ID_ANY, "0.1"),
                      'n_iter_no_change': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'tol': wx.TextCtrl(self, wx.ID_ANY, "1e-4")}

        self.defaults = {'loss': 'ls',
                         'learning_rate': 0.1,
                         'n_estimators': 100,
                         'subsample': 1.0,
                         'criterion': 'friedman_mse',
                         'max_depth': 3,
                         'min_samples_split': 2,
                         'min_samples_leaf': 1,
                         'min_weight_fraction_leaf': 0,
                         'max_features': None,
                         'alpha': 0.9,
                         'max_leaf_nodes': None,
                         'min_impurity_decrease': 0,
                         'init': 'DummyEstimator',
                         'random_state': None,
                         'presort': 'auto',
                         'validation_fraction': 0.1,
                         'n_iter_no_change': None,
                         'tol': float('1e-4')}

        self.getters = {'loss': self.to_str,
                        'learning_rate': self.to_float,
                        'n_estimators': self.to_int,
                        'subsample': self.to_float,
                        'criterion': self.to_str,
                        'max_depth': self.to_int,
                        'min_samples_split': self.to_int_or_float,
                        'min_samples_leaf': self.to_int_or_float,
                        'min_weight_fraction_leaf': self.to_float,
                        'max_features': self.to_int_float_string_or_none,
                        'alpha': self.to_float,
                        'max_leaf_nodes': self.to_int_or_none,
                        'min_impurity_decrease': self.to_float,
                        'init': self.to_str,
                        'random_state': self.to_int_or_none,
                        'presort': self.to_bool_or_str,
                        'validation_fraction': self.to_float,
                        'n_iter_no_change': self.to_int_or_none,
                        'tol': self.to_float}

        self.run()


class DecisionTreeFrame(MachineLearningFrame):
    def __init__(self, main_app_frame, data, include_test_data=True):
        MachineLearningFrame.__init__(self, main_app_frame, data, 'Decision Tree', include_test_data=include_test_data)

        self.input = {'criterion': wx.ComboBox(self, wx.ID_ANY, choices=["mse", "friedman_mse", "mae"],
                                               style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'splitter': wx.ComboBox(self, wx.ID_ANY, choices=["best", "random"],
                                              style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'max_depth': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'min_samples_split': wx.TextCtrl(self, wx.ID_ANY, "2"),
                      'min_samples_leaf': wx.TextCtrl(self, wx.ID_ANY, "1"),
                      'min_weight_fraction_leaf': wx.TextCtrl(self, wx.ID_ANY, "0"),
                      'max_features': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'max_leaf_nodes': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'min_impurity_decrease': wx.TextCtrl(self, wx.ID_ANY, "0"),
                      'random_state': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'presort': wx.ComboBox(self, wx.ID_ANY, choices=["True", "False"],
                                             style=wx.CB_DROPDOWN | wx.CB_READONLY)}

        self.defaults = {'criterion': 'mse',
                         'splitter': 'best',
                         'max_depth': None,
                         'min_samples_split': 2,
                         'min_samples_leaf': 1,
                         'min_weight_fraction_leaf': 0.,
                         'max_features': None,
                         'max_leaf_nodes': None,
                         'min_impurity_decrease': 0.,
                         'random_state': None,
                         'presort': False}

        self.getters = {'criterion': self.to_str,
                        'splitter': self.to_str,
                        'max_depth': self.to_int_or_none,
                        'min_samples_split': self.to_int_or_float,
                        'min_samples_leaf': self.to_int_or_float,
                        'min_weight_fraction_leaf': self.to_float,
                        'max_features': self.to_int_float_string_or_none,
                        'max_leaf_nodes': self.to_int_or_none,
                        'min_impurity_decrease': self.to_float,
                        'random_state': self.to_int_or_none,
                        'presort': self.to_bool}

        self.run()


class SupportVectorRegressionFrame(MachineLearningFrame):
    def __init__(self, main_app_frame, data, include_test_data=True):
        MachineLearningFrame.__init__(self, main_app_frame, data, 'Support Vector Machine',
                                      include_test_data=include_test_data)

        self.input = {'kernel': wx.ComboBox(self, wx.ID_ANY, "rbf",
                                            choices=["linear", "poly", "rbf", "sigmoid", "precomputed"],
                                            style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'degree': wx.TextCtrl(self, wx.ID_ANY, "3"),
                      'gamma': wx.TextCtrl(self, wx.ID_ANY, "auto"),
                      'coef0': wx.TextCtrl(self, wx.ID_ANY, "0.0"),
                      'tol': wx.TextCtrl(self, wx.ID_ANY, "1e-3"),
                      'C': wx.TextCtrl(self, wx.ID_ANY, "1.0"),
                      'epsilon': wx.TextCtrl(self, wx.ID_ANY, "0.1"),
                      'shrinking': wx.ComboBox(self, wx.ID_ANY, "True",
                                               choices=["True", "False"],
                                               style=wx.CB_DROPDOWN | wx.CB_READONLY),
                      'cache_size': wx.TextCtrl(self, wx.ID_ANY, "None"),
                      'max_iter': wx.TextCtrl(self, wx.ID_ANY, "-1")}

        self.defaults = {'kernel': 'rbf',
                         'degree': 3,
                         'gamma': 'scale',
                         'coef0': 0.0,
                         'tol': 0.001,
                         'C': 1.0,
                         'epsilon': 0.1,
                         'shrinking': True,
                         'cache_size': None,
                         'max_iter': -1}

        self.getters = {'kernel': self.to_str,
                        'degree': self.to_int,
                        'gamma': self.to_float_or_str,
                        'coef0': self.to_float,
                        'tol': self.to_float,
                        'C': self.to_float,
                        'epsilon': self.to_float,
                        'shrinking': self.to_bool,
                        'cache_size': self.to_float_or_none,
                        'max_iter': self.to_int}

        self.button_importance.Disable()

        self.run()


RF_TOOL_TIPS = {'n_estimators': "int\nThe number of trees in the forest.",
                'criterion': "The function to measure the quality of a split. Supported criteria are "
                             "“mse” for the mean squared error, which is equal to variance reduction as"
                             " feature selection criterion, and “mae” for the mean absolute error.",
                'max_depth': "int, None\nThe maximum depth of the tree. If None, then nodes are expanded until all "
                             "leaves are pure or until all leaves contain less than min_samples_split samples.",
                'min_samples_split': "int, float\nThe minimum number of samples required to split an "
                                     "internal node:\n• If int, then consider min_samples_split as the "
                                     "minimum number.\n• If float, then min_samples_split is a fraction "
                                     "and ceil(min_samples_split * n_samples) are the minimum number "
                                     "of samples for each split.",
                'min_samples_leaf': "int, float\nThe minimum number of samples required to be at a leaf"
                                    " node. A split point at any depth will only be considered if it "
                                    "leaves at least min_samples_leaf training samples in each of the "
                                    "left and right branches. This may have the effect of smoothing "
                                    "the model, especially in regression.\n• If int, then consider min_"
                                    "samples_leaf as the minimum number.\n• If float, then min_samples_"
                                    "leaf is a fraction and ceil(min_samples_leaf * n_samples) are the "
                                    "minimum number of samples for each node.",
                'min_weight_fraction_leaf': "float\nThe minimum weighted fraction of the sum total of "
                                            "weights (of all the input samples) required to be at a leaf"
                                            " node. Samples have equal weight when sample_weight is not"
                                            " provided.",
                'max_features': "int, float, string, or None\nThe number of features to consider when "
                                "looking for the best split:\n• If int, then consider max_features "
                                "features at each split.\n• If float, then max_features is a fraction"
                                " and int(max_features * n_features) features are considered at each "
                                "split.\n• If “auto”, then max_features=n_features.\n• If “sqrt”, "
                                "then max_features=sqrt(n_features).\n• If “log2”, then max_"
                                "features=log2(n_features).\n• If None, then max_features=n_features.",
                'max_leaf_nodes': "int or None\nGrow a tree with max_leaf_nodes in best-first fashion. "
                                  "Best nodes are defined as relative reduction in impurity. If None "
                                  "then unlimited number of leaf nodes.",
                'min_impurity_decrease': "float\nA node will be split if this split induces a decrease "
                                         "of the impurity greater than or equal to this value.\nThe "
                                         "weighted impurity decrease equation is the following:\n"
                                         "N_t / N * (impurity - N_t_R / N_t * right_impurity\n"
                                         "                    - N_t_L / N_t * left_impurity)\n"
                                         "where N is the total number of samples, N_t is the number of "
                                         "samples at the current node, N_t_L is the number of samples "
                                         "in the left child, and N_t_R is the number of samples in the "
                                         "right child.\nN, N_t, N_t_R and N_t_L all refer to the "
                                         "weighted sum, if sample_weight is passed.",
                'bootstrap': "Whether bootstrap samples are used when building trees. If False, the "
                             "whole datset is used to build each tree.",
                'oob_score': "whether to use out-of-bag samples to estimate the R^2 on unseen data.",
                'n_jobs': "int or None\nThe number of jobs to run in parallel for both fit and "
                          "predict. None` means 1 unless in a joblib.parallel_backend context. -1 "
                          "means using all processors.",
                'random_state': "int or None\nIf int, random_state is the seed used by the random "
                                "number generator; If None, the random number generator is the "
                                "RandomState instance used by np.random."}

GB_TOOL_TIPS = {'loss': "loss function to be optimized. ‘ls’ refers to least squares regression. "
                        "‘lad’ (least absolute deviation) is a highly robust loss function solely "
                        "based on order information of the input variables. ‘huber’ is a combination "
                        "of the two. ‘quantile’ allows quantile regression (use alpha to specify the "
                        "quantile).",
                'learning_rate': "float\nlearning rate shrinks the contribution of each tree by "
                                 "learning_rate. There is a trade-off between learning_rate and n_estimators.",
                'n_estimators': "int\nThe number of boosting stages to perform. Gradient boosting is "
                                "fairly robust to over-fitting so a large number usually results in "
                                "better performance.",
                'subsample': "float\nThe fraction of samples to be used for fitting the individual "
                             "base learners. If smaller than 1.0 this results in Stochastic Gradient "
                             "Boosting. subsample interacts with the parameter n_estimators. Choosing "
                             "subsample < 1.0 leads to a reduction of variance and an increase in bias.",
                'criterion': "The function to measure the quality of a split. Supported criteria are"
                             " “friedman_mse” for the mean squared error with improvement score by "
                             "Friedman, “mse” for mean squared error, and “mae” for the mean absolute"
                             " error. The default value of “friedman_mse” is generally the best as it "
                             "can provide a better approximation in some cases.",
                'max_depth': "int\nmaximum depth of the individual regression estimators. The maximum "
                             "depth limits the number of nodes in the tree. Tune this parameter for "
                             "best performance; the best value depends on the interaction of the input"
                             " variables.",
                'min_samples_split': "int, float\nThe minimum number of samples required to split an "
                                     "internal node:\n• If int, then consider min_samples_split as "
                                     "the minimum number.\n• If float, then min_samples_split is a "
                                     "fraction and ceil(min_samples_split * n_samples) are the minimum"
                                     " number of samples for each split.",
                'min_samples_leaf': "int, float\nThe minimum number of samples required to be at a "
                                    "leaf node. A split point at any depth will only be considered if "
                                    "it leaves at least min_samples_leaf training samples in each of "
                                    "the left and right branches. This may have the effect of "
                                    "smoothing the model, especially in regression.\n• If int, then "
                                    "consider min_samples_leaf as the minimum number.\n• If float, "
                                    "then min_samples_leaf is a fraction and ceil(min_samples_leaf * "
                                    "n_samples) are the minimum number of samples for each node.",
                'min_weight_fraction_leaf': "float\nThe minimum weighted fraction of the sum total of "
                                            "weights (of all the input samples) required to be at a "
                                            "leaf node. Samples have equal weight when sample_weight is"
                                            " not provided.",
                'max_features': "int, float, string, or None\nThe number of features to consider when"
                                " looking for the best split:\n• If int, then consider max_features"
                                " features at each split.\n• If float, then max_features is a "
                                "fraction and int(max_features * n_features) features are considered "
                                "at each split.\n• If “auto”, then max_features=n_features.\n• If "
                                "“sqrt”, then max_features=sqrt(n_features).\n• If “log2”, then max"
                                "_features=log2(n_features).\n• If None, then max_features=n_features."
                                "\nChoosing max_features < n_features leads to a reduction of variance"
                                " and an increase in bias.",
                'alpha': "float\nThe alpha-quantile of the huber loss function and the quantile loss "
                         "function. Only if loss='huber' or loss='quantile'.",
                'max_leaf_nodes': "int or None\nGrow a tree with max_leaf_nodes in best-first fashion. "
                                  "Best nodes are defined as relative reduction in impurity. If None "
                                  "then unlimited number of leaf nodes.",
                'min_impurity_decrease': "float\nA node will be split if this split induces a decrease "
                                         "of the impurity greater than or equal to this value.\nThe "
                                         "weighted impurity decrease equation is the following:\n"
                                         "N_t / N * (impurity - N_t_R / N_t * right_impurity\n"
                                         "                    - N_t_L / N_t * left_impurity)\nwhere N is"
                                         " the total number of samples, N_t is the number of samples at"
                                         " the current node, N_t_L is the number of samples in the left"
                                         " child, and N_t_R is the number of samples in the right child."
                                         "\nN, N_t, N_t_R and N_t_L all refer to the weighted sum, if "
                                         "sample_weight is passed.",
                'init': "If ‘zero’, the initial raw predictions are set to zero. By default a "
                        "DummyEstimator is used, predicting either the average target value "
                        "(for loss=’ls’), or a quantile for the other losses.",
                'random_state': "int or None\nIf int, random_state is the seed used by the random number"
                                " generator; If None, the random number generator is the RandomState "
                                "instance used by np.random.",
                'presort': "Whether to presort the data to speed up the finding of best splits in "
                           "fitting. Auto mode by default will use presorting on dense data and default "
                           "to normal sorting on sparse data. Setting presort to true on sparse data "
                           "will raise an error.",
                'validation_fraction': "float\nThe proportion of training data to set aside as "
                                       "validation set for early stopping. Must be between 0 and 1. "
                                       "Only used if n_iter_no_change is set to an integer.",
                'n_iter_no_change': "int or None\nn_iter_no_change is used to decide if early stopping "
                                    "will be used to terminate training when validation score is not "
                                    "improving. By default it is set to None to disable early stopping."
                                    " If set to a number, it will set aside validation_fraction size of"
                                    " the training data as validation and terminate training when "
                                    "validation score is not improving in all of the previous "
                                    "n_iter_no_change numbers of iterations.",
                'tol': "float\nTolerance for the early stopping. When the loss is not improving by at "
                        "least tol for n_iter_no_change iterations (if set to a number), the training "
                        "stops."}

DT_TOOL_TIPS = {'criterion': "The function to measure the quality of a split. Supported criteria are "
                             "“mse” for the mean squared error, which is equal to variance reduction "
                             "as feature selection criterion and minimizes the L2 loss using the mean "
                             "of each terminal node, “friedman_mse”, which uses mean squared error "
                             "with Friedman’s improvement score for potential splits, and “mae” for "
                             "the mean absolute error, which minimizes the L1 loss using the median of"
                             " each terminal node.",
                'splitter': "The strategy used to choose the split at each node. Supported strategies "
                            "are “best” to choose the best split and “random” to choose the best "
                            "random split.",
                'max_depth': "int, None\nThe maximum depth of the tree. If None, then nodes are expanded"
                             " until all leaves are pure or until all leaves contain less than "
                             "min_samples_split samples.",
                'min_samples_split': "int, float\nThe minimum number of samples required to split an "
                                     "internal node:\n• If int, then consider min_samples_split as the "
                                     "minimum number.\n• If float, then min_samples_split is a fraction"
                                     " and ceil(min_samples_split * n_samples) are the minimum number "
                                     "of samples for each split.",
                'min_samples_leaf': "int, float\nThe minimum number of samples required to be at a leaf"
                                    " node. A split point at any depth will only be considered if it "
                                    "leaves at least min_samples_leaf training samples in each of the "
                                    "left and right branches. This may have the effect of smoothing the"
                                    " model, especially in regression.\n• If int, then consider "
                                    "min_samples_leaf as the minimum number.\n• If float, then "
                                    "min_samples_leaf is a fraction and ceil(min_samples_leaf * "
                                    "n_samples) are the minimum number of samples for each node.",
                'min_weight_fraction_leaf': "float\nThe minimum weighted fraction of the sum total of "
                                            "weights (of all the input samples) required to be at a leaf"
                                            " node. Samples have equal weight when sample_weight is not"
                                            " provided.",
                'max_features': "int, float, string, or None\nThe number of features to consider when"
                                " looking for the best split:\n• If int, then consider max_features "
                                "features at each split.\n• If float, then max_features is a fraction"
                                " and int(max_features * n_features) features are considered at each "
                                "split.\n• If “auto”, then max_features=n_features.\n• If “sqrt”, "
                                "then max_features=sqrt(n_features).\n• If “log2”, "
                                "then max_features=log2(n_features).\n• If None, then "
                                "max_features=n_features.",
                'max_leaf_nodes': "int or None\nGrow a tree with max_leaf_nodes in best-first fashion. "
                                  "Best nodes are defined as relative reduction in impurity. If None "
                                  "then unlimited number of leaf nodes.",
                'min_impurity_decrease': "float\nA node will be split if this split induces a decrease "
                                         "of the impurity greater than or equal to this value.\nThe "
                                         "weighted impurity decrease equation is the following:\n"
                                         "N_t / N * (impurity - N_t_R / N_t * right_impurity\n"
                                         "                    - N_t_L / N_t * left_impurity)\nwhere N is"
                                         " the total number of samples, N_t is the number of samples at"
                                         " the current node, N_t_L is the number of samples in the left"
                                         " child, and N_t_R is the number of samples in the right child."
                                         "\nN, N_t, N_t_R and N_t_L all refer to the weighted sum, if "
                                         "sample_weight is passed.",
                'random_state': "int or None\nIf int, random_state is the seed used by the random number"
                                " generator; If None, the random number generator is the RandomState "
                                "instance used by np.random.",
                'presort': "Whether to presort the data to speed up the finding of best splits in "
                           "fitting. For the default settings of a decision tree on large datasets, "
                           "setting this to true may slow down the training process. When using either "
                           "a smaller dataset or a restricted depth, this may speed up the training."}

SVR_TOOL_TIPS = {'kernel': "string\n"
                           "Specifies the kernel type to be used in the algorithm. It must be one of ‘linear’, ‘poly’, "
                           "‘rbf’, ‘sigmoid’, ‘precomputed’ or a callable. If none is given, ‘rbf’ will be used. If "
                           "a callable is given it is used to precompute the kernel matrix.",
                 'degree': "int\n"
                           "Degree of the polynomial kernel function (‘poly’). Ignored by all other kernels.",
                 'gamma': "float\n"
                          "Kernel coefficient for ‘rbf’, ‘poly’ and ‘sigmoid’.\n\n"
                          "Current default is ‘auto’ which uses 1 / n_features, if gamma='scale' is passed then it "
                          "uses 1 / (n_features * X.var()) as value of gamma. The current default of gamma, ‘auto’, "
                          "will change to ‘scale’ in version 0.22. ‘auto_deprecated’, a deprecated version of ‘auto’ "
                          "is used as a default indicating that no explicit value of gamma was passed.",
                 'coef0': "float\n"
                          "Independent term in kernel function. It is only significant in ‘poly’ and ‘sigmoid’",
                 'tol': "float\n"
                        "Tolerance for stopping criterion.",
                 'C': "float\n"
                      "Penalty parameter C of the error term.",
                 'epsilon': "float\n"
                            "Epsilon in the epsilon-SVR model. It specifies the epsilon-tube within which no penalty is"
                            " associated in the training loss function with points predicted within a distance epsilon"
                            " from the actual value.",
                 'shrinking': "boolean\n"
                              "Whether to use the shrinking heuristic",
                 'cache_size': "float\n"
                               "Specify the size of the kernel cache (in MB)",
                 'max_iter': "int\n"
                             "Hard limit on iterations within solver, or -1 for no limit."}

DATA_SPLIT_TOOL_TIPS = {'test_size': "float, int, or None\n"
                                     "If float, should be between 0.0 and 1.0 and represent the proportion of the "
                                     "dataset to include in the test split. If int, represents the absolute number of "
                                     "test samples. If None, the value is set to the complement of the train size. If "
                                     "train_size is also None, it will be set to 0.25.",
                        'train_size': "float, int, or None\n"
                                      "If float, should be between 0.0 and 1.0 and represent the proportion of the "
                                      "dataset to include in the train split. If int, represents the absolute number "
                                      "of train samples. If None, the value is automatically set to the complement of "
                                      "the test size.",
                        'random_state': "int or None\n"
                                        "If int, random_state is the seed used by the random number generator; If None,"
                                        " the random number generator is the RandomState instance used by np.random.",
                        'shuffle': "boolean\n"
                                   "Whether or not to shuffle the data before splitting. If shuffle=False then stratify"
                                   " must be None."}


ALGORITHMS = {'Random Forest': {'regressor': RandomForestRegressor,
                                'tool_tips': RF_TOOL_TIPS,
                                'frame': RandomForestFrame},
              'Support Vector Machine': {'regressor': SVR,
                                         'tool_tips': SVR_TOOL_TIPS,
                                         'frame': SupportVectorRegressionFrame},
              'Gradient Boosting': {'regressor': GradientBoostingRegressor,
                                    'tool_tips': GB_TOOL_TIPS,
                                    'frame': GradientBoostingFrame},
              'Decision Tree': {'regressor': DecisionTreeRegressor,
                                'tool_tips': DT_TOOL_TIPS,
                                'frame': DecisionTreeFrame}}


class MachineLearningPlotData:
    def __init__(self, X, y, reg, do_training=True, **kwargs):
        self.reg = reg
        self.split_args = kwargs

        indices = list(range(len(y)))

        # split the data for training and testing
        split_data = train_test_split(X, indices, **kwargs)
        self.X = {'data': X, 'train': split_data[0], 'test': split_data[1]}
        self.indices = {'data': indices, 'train': split_data[2], 'test': split_data[3]}
        self.y = {'data': y, 'train': [y[i] for i in split_data[2]], 'test': [y[i] for i in split_data[3]]}
        self.x = {key: [i + 1 for i in range(len(data))] for key, data in self.y.items()}

        # Train model, then calculate predictions, residuals, and mse
        if do_training:
            self.reg.fit(self.X['train'], self.y['train'])
        self.predictions = {key: self.get_prediction(key) for key in self.y.keys()}
        self.residuals = {key: self.get_residual(key) for key in self.y.keys()}
        self.mse = {key: self.get_mse(key) for key in self.y.keys()}

    def get_prediction(self, key):
        return self.reg.predict(self.X[key])

    def get_mse(self, key):
        return np.mean(np.square(np.subtract(self.predictions[key], self.y[key])))

    def get_residual(self, key):
        return np.subtract(self.y[key], self.reg.predict(self.X[key]))

    @property
    def feature_importances(self):
        if hasattr(self.reg, 'feature_importances_'):
            return self.reg.feature_importances_
        return None


class FeatureImportanceFrame(wx.Frame):
    def __init__(self, options, x_variables, feature_importances, frame_title, plot_title):
        wx.Frame.__init__(self, None)

        self.title = frame_title

        self.plot = PlotFeatureImportance(self, options, x_variables, feature_importances, plot_title)

        self.set_properties()
        self.do_layout()

        self.Bind(wx.EVT_SIZE, self.on_resize)

    def set_properties(self):
        self.SetTitle(self.title)
        self.SetMinSize(get_window_size(0.35, 0.8))

    def do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.HORIZONTAL)
        sizer_wrapper.Add(self.plot.layout, 1, wx.EXPAND, 0)

        set_msw_background_color(self)  # If windows, change the background color
        set_frame_icon(self)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def redraw_plot(self):
        self.plot.redraw_plot()

    def on_resize(self, *evt):
        try:
            self.Refresh()
            self.Layout()
            wx.CallAfter(self.redraw_plot)
        except RuntimeError:
            pass


class MachineLearningModelViewer:
    def __init__(self, parent, group_data, group, options, mvr=None):
        self.parent = parent
        self.group_data = group_data
        self.group = group
        self.stats_data = group_data[group]['stats_data']
        self.options = options

        self.file_path = self.file_select_dlg()

        if self.file_path:

            self.__load_mlr_file()
            try:
                if self.is_valid:
                    self.__set_X_and_y_data()

                    if mvr:
                        self.mvr = mvr
                    else:
                        self.mvr = MultiVariableRegression(self.X, self.y)
                    self.multi_var_pred = self.mvr.predictions

                    data_keys = ['X', 'y', 'x_variables', 'y_variable', 'multi_var_pred', 'options', 'mrn', 'study_date', 'uid']
                    data = {key: getattr(self, key) for key in data_keys}
                    frame = ALGORITHMS[self.title]['frame']
                    self.ml_frame = frame(data, include_test_data=False)
                    self.__load_model()

                    set_msw_background_color(self.ml_frame)
                    set_frame_icon(self.ml_frame)

                    self.ml_frame.run()
                else:
                    if self.stats_data is None:
                        msg = 'No data has been queried for Group %s.' % group
                    elif not self.is_mlr:
                        msg = 'Selected file is not a valid machine learning regression save file.'
                    elif not self.stats_data_has_y:
                        msg = "The model's dependent variable is not found in your queried data:\n%s" % self.y_variable
                    elif self.missing_x_variables:
                        msg = 'Your queried data is missing the following independent variables:\n%s' % \
                              ', '.join(self.missing_x_variables)
                    else:
                        msg = 'Unknown error.'

                    wx.MessageBox(msg, 'Model Loading Error', wx.OK | wx.OK_DEFAULT | wx.ICON_WARNING)
            except Exception as e:
                msg = str(e)
                wx.MessageBox(msg, 'Model Loading Error', wx.OK | wx.OK_DEFAULT | wx.ICON_WARNING)

    def file_select_dlg(self):
        with wx.FileDialog(self.parent, "Load a machine learning regression model", wildcard='*.mlr',
                           style=wx.FD_FILE_MUST_EXIST | wx.FD_OPEN) as dlg:
            dlg.SetDirectory(MODELS_DIR)
            if dlg.ShowModal() == wx.ID_OK:
                return dlg.GetPath()

    def __load_mlr_file(self):
        self.loaded_data = load_object_from_file(self.file_path)

        self.y_variable = self.loaded_data['y_variable']
        self.regression = self.loaded_data['regression']
        self.regressor = self.loaded_data['regressor']
        self.tool_tips = self.loaded_data['tool_tips']
        self.x_variables = self.loaded_data['x_variables']
        self.title = self.loaded_data['title']
        self.input_parameters = self.loaded_data['input_parameters']
        self.data_split = self.loaded_data['data_split']
        self.version = self.loaded_data['version']

    def __load_model(self):
        self.ml_frame.reg = self.regression
        self.ml_frame.set_input_parameters(self.input_parameters)
        self.ml_frame.set_data_split_parameters(self.data_split)
        self.__disable_input()

    def __disable_input(self):
        for variable, input_obj in self.ml_frame.input.items():
            input_obj.Disable()
        for variable, input_obj in self.ml_frame.data_split_input.items():
            input_obj.Disable()
        self.ml_frame.button_calculate.Disable()
        self.ml_frame.button_save_model.Disable()

    def __set_X_and_y_data(self):
        self.X, self.y, self.mrn, self.uid, self.study_date = \
            self.stats_data.get_X_and_y(self.y_variable, self.x_variables, include_patient_info=True)

    @property
    def is_mlr(self):
        return 'title' in list(self.loaded_data) \
               and self.loaded_data['title'] in list(ALGORITHMS)

    @property
    def is_valid(self):
        return self.stats_data is not None and self.is_mlr and not self.missing_x_variables and self.stats_data_has_y

    @property
    def missing_x_variables(self):
        return [x for x in self.x_variables if x not in list(self.stats_data.data)]

    @property
    def stats_data_has_y(self):
        return self.y_variable in list(self.stats_data.data)

    def apply_plot_options(self):
        self.ml_frame.plot.apply_options()
        self.ml_frame.redraw_plot()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.queried_data.py
"""
Class for viewing SQL table data of the current query
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from dvha.models.data_table import DataTable
from dvha.dialogs.export import save_data_to_file
from dvha.tools.utilities import get_window_size


class QueriedDataFrame(wx.Frame):
    """
    Generate a simple table to view data of the current query for a specified SQL table
    """
    def __init__(self, data_obj, columns, data_key, menu, menu_item_id):
        """
        :param data_obj: object containing data to be viewed for each group
        :type data_obj: dict
        :param columns: columns to be displayed in table
        :type columns: list
        :param data_key: either 'DVHs', 'Plans', 'Beams', 'Rxs', or 'StatsData'
        :type data_key: str
        :param menu: a link to the main app menu, used to toggle Show/Hide status
        :type menu: Menu
        :param menu_item_id: the ID of the menu item associated with the specified data_obj
        """
        wx.Frame.__init__(self, None, title='%s Data' % data_key[0:-1])

        self.data = data_obj
        self.columns = columns
        self.sql_table = data_key
        self.menu = menu
        self.menu_item_id = menu_item_id

        self.list_ctrl = wx.ListCtrl(self, wx.ID_ANY,
                                     style=wx.BORDER_SUNKEN | wx.LC_HRULES | wx.LC_REPORT | wx.LC_VRULES)

        self.button_export = wx.Button(self, wx.ID_ANY, "Export to CSV")
        self.radio_button_query_group = wx.RadioBox(self, wx.ID_ANY, 'Query Group', choices=['1', '2'])

        self.data_table = DataTable(self.list_ctrl)
        self.data_table.set_data(self.table_data, self.columns)

        if not self.data[2]:
            self.radio_button_query_group.Disable()

        self.__do_bind()
        self.__set_properties()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.SetSize(get_window_size(0.714, 0.762))

    def __do_bind(self):
        self.Bind(wx.EVT_BUTTON, self.on_export, id=self.button_export.GetId())
        self.Bind(wx.EVT_LIST_COL_CLICK, self.sort_table, self.list_ctrl)
        self.Bind(wx.EVT_CLOSE, self.on_close)
        self.Bind(wx.EVT_RADIOBOX, self.on_group_select, id=self.radio_button_query_group.GetId())

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_widgets = wx.BoxSizer(wx.HORIZONTAL)
        sizer_button = wx.BoxSizer(wx.HORIZONTAL)
        sizer_button.Add(self.button_export, 0, wx.TOP, 15)
        sizer_widgets.Add(sizer_button, 0, wx.ALL, 10)
        sizer_widgets.Add(self.radio_button_query_group, 0, wx.ALL, 10)
        sizer_wrapper.Add(sizer_widgets, 0, 0, 0)
        sizer_wrapper.Add(self.list_ctrl, 1, wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND, 10)
        self.SetSizer(sizer_wrapper)
        self.Center()

    @property
    def table_data(self):
        return {column: getattr(self.selected_data, column) for column in self.columns}

    @property
    def selected_group(self):
        return self.radio_button_query_group.GetSelection() + 1

    @property
    def selected_data(self):
        return self.data[self.selected_group]

    def run(self):
        self.toggle_data_menu_item()
        self.Show()

    def on_close(self, *args):
        self.toggle_data_menu_item()
        self.Destroy()

    def on_export(self, *args):
        save_data_to_file(self, "Export Group %s %s to CSV" % (self.selected_group, self.sql_table),
                          self.data_table.get_csv())

    def toggle_data_menu_item(self):
        short_cut = ['DVHs', 'Plans', 'Rxs', 'Beams'].index(self.sql_table) + 1
        show_hide = ['Show', 'Hide']['Show' in self.menu.GetLabel(self.menu_item_id)]
        self.menu.SetLabel(self.menu_item_id, '%s %s\tCtrl+%s' % (show_hide, self.sql_table, short_cut))

    def sort_table(self, evt):
        self.data_table.sort_table(evt)

    def on_group_select(self, *evt):
        self.data_table.clear()
        self.data_table.set_data(self.table_data, self.columns)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# paths.py
"""
A collection of directories and paths updated with the script directory and user's home folder for the OS
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

from os.path import join, dirname, expanduser

SCRIPT_DIR = dirname(__file__)
ICONS_DIR = join(SCRIPT_DIR, 'icons')
WIN_APP_ICON = join(ICONS_DIR, 'dvha.ico')
LOGO_PATH = join(SCRIPT_DIR, 'logo.png')
APPS_DIR = join(expanduser('~'), 'Apps')
APP_DIR = join(APPS_DIR, 'dvh_analytics')
PREF_DIR = join(APP_DIR, 'preferences')
DATA_DIR = join(APP_DIR, 'data')
INBOX_DIR = join(DATA_DIR, 'inbox')
IMPORTED_DIR = join(DATA_DIR, 'imported')
REVIEW_DIR = join(DATA_DIR, 'review')
BACKUP_DIR = join(DATA_DIR, 'backup')
TEMP_DIR = join(DATA_DIR, 'temp')
MODELS_DIR = join(DATA_DIR, 'models')
PROTOCOL_DIR = join(DATA_DIR, 'protocols')
PROTOCOL_DEFAULT_DIR = join(SCRIPT_DIR, 'protocols')
DIRECTORIES = {key[:-4]: value for key, value in locals().items() if key.endswith('_DIR')}

OPTIONS_PATH = join(PREF_DIR, '.options')
OPTIONS_CHECKSUM_PATH = join(PREF_DIR, '.options_checksum')
IMPORT_SETTINGS_PATH = join(PREF_DIR, 'import_settings.txt')
SQL_CNF_PATH = join(PREF_DIR, 'sql_connection.cnf')
LICENSE_PATH = join(SCRIPT_DIR, 'LICENSE.txt')
CREATE_PGSQL_TABLES = join(SCRIPT_DIR, 'db', 'create_tables.sql')
CREATE_SQLITE_TABLES = join(SCRIPT_DIR, 'db', 'create_tables_sqlite.sql')
TG263_CSV = join(SCRIPT_DIR, 'tools', 'TG263_Nomenclature_Worksheet_20170815.csv')
PIP_LIST_PATH = join(SCRIPT_DIR, 'tools', 'pip_list')

ICONS = {'Open': "iconfinder_Open_1493293.png",
         'Close': "iconfinder_Close_1493281.png",
         'Save': "iconfinder_Save_1493294.png",
         'Print': "iconfinder_Print_1493286.png",
         'Export': "iconfinder_csv_file_database_extension_data_3876336.png",
         'Import': "iconfinder_import_4168538.png",
         'Settings': "iconfinder_Settings_1493289.png",
         'Database': "iconfinder_data_115746_black_edit.png",
         'ROI Map': "iconfinder_icon-map_211858.png",
         'ok-green': "iconfinder_ok-green_53916.png",
         'ko-red': "iconfinder_ko-red_53948.png",
         'custom_Y': "icon_custom_Y.png",
         'custom_X': "icon_custom_X.png",
         'rtplan': "iconfinder_Clipboard-Plan_379537_zoom.png",
         'rtstruct': "iconfinder_Education-Filled_7_3672892.png",
         'rtdose': "iconfinder_package-supported_24220.png",
         'other': "error.png",
         'studies': "iconfinder_User_Customers_1218712.png",
         'study': "iconfinder_Travel-Filled-07_3671983.png",
         'plan': "iconfinder_Clipboard-Plan_379537_zoom.png",
         'patient': "iconfinder_User_Yuppie_3_1218716.png",
         'Image': "iconfinder_m-52_4230522.png"}
for key, value in ICONS.items():
    ICONS[key] = join(ICONS_DIR, value)


def parse_settings_file(abs_file_path):
    with open(abs_file_path, 'r') as document:
        settings = {}
        for line in document:
            line = line.split()
            if not line:
                continue
            if len(line) > 1:
                settings[line[0]] = line[1:][0]
                # Convert strings to boolean
                if line[1:][0].lower() == 'true':
                    settings[line[0]] = True
                elif line[1:][0].lower() == 'false':
                    settings[line[0]] = False
            else:
                settings[line[0]] = ''
    return settings



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# models.time_series.py
"""
Class for viewing and editing the roi map, and updating the database with changes
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from dvha.dialogs.export import save_data_to_file
from dvha.models.plot import PlotCorrelation
from dvha.dialogs.main import SelectFromListDialog
from dvha.tools.errors import ErrorDialog
from dvha.tools.utilities import get_window_size


class CorrelationFrame:
    """
    Object to be passed into notebook panel for the Time Series tab
    """
    def __init__(self, parent, group_data, options):
        """
        :param parent:  notebook panel in main view
        :type parent: Panel
        :param group_data: dvh, table, and stats_data
        :type group_data: dict
        :param options: user options containing visual preferences
        :type options: Options
        """
        self.parent = parent
        self.options = options
        self.stats_data = group_data[1]['stats_data']
        self.stats_data_2 = group_data[2]['stats_data']

        self.selections = options.CORRELATION_MATRIX_VARS

        self.__define_gui_objects()
        self.__do_bind()
        self.__do_layout()

    def __define_gui_objects(self):
        self.button_var_select = wx.Button(self.parent, wx.ID_ANY, 'Select Variables')
        self.button_var_default = wx.Button(self.parent, wx.ID_ANY, 'Default Variables')
        self.button_export_csv = wx.Button(self.parent, wx.ID_ANY, 'Export')
        self.button_save_html = wx.Button(self.parent, wx.ID_ANY, 'Save HTML')
        self.button_save_svg = wx.Button(self.parent, wx.ID_ANY, 'Save SVG')
        self.plot = PlotCorrelation(self.parent, self.options)

    def __do_bind(self):
        self.parent.Bind(wx.EVT_BUTTON, self.on_var_select, id=self.button_var_select.GetId())
        self.parent.Bind(wx.EVT_BUTTON, self.on_var_default, id=self.button_var_default.GetId())
        self.parent.Bind(wx.EVT_BUTTON, self.export_csv, id=self.button_export_csv.GetId())
        self.parent.Bind(wx.EVT_BUTTON, self.save_html, id=self.button_save_html.GetId())
        self.parent.Bind(wx.EVT_BUTTON, self.save_svg, id=self.button_save_svg.GetId())

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_plot = wx.BoxSizer(wx.HORIZONTAL)

        sizer_buttons.Add(self.button_var_select, 0, wx.RIGHT, 5)
        sizer_buttons.Add(self.button_var_default, 0, wx.RIGHT, 5)
        sizer_buttons.Add(self.button_export_csv, 0, wx.RIGHT, 5)
        sizer_buttons.Add(self.button_save_html, 0, wx.RIGHT, 5)
        sizer_buttons.Add(self.button_save_svg, 0, wx.RIGHT, 5)

        sizer_plot.Add(self.plot.layout, 1, wx.EXPAND, 0)

        sizer_wrapper.Add(sizer_buttons, 0, wx.BOTTOM, 5)
        sizer_wrapper.Add(sizer_plot, 1, wx.EXPAND, 0)

        self.layout = sizer_wrapper

    def on_var_select(self, evt):
        categories = [c for c in list(self.stats_data.data) if 'date' not in c.lower()]
        categories.sort()
        size = get_window_size(1, 0.8)
        dlg = SelectFromListDialog("Correlation Matrix", "Variables", categories,
                                   size=(350, size[1]), column_width=300, selections=self.selections)
        res = dlg.ShowModal()

        if res == wx.ID_OK:
            self.selections = dlg.selected_values
            self.update_plot_data()
        dlg.Destroy()

    def on_var_default(self, evt):
        self.selections = self.options.CORRELATION_MATRIX_VARS
        self.update_plot_data()

    def set_data(self, group_data):
        self.stats_data = group_data[1]['stats_data']
        self.stats_data_2 = group_data[2]['stats_data']
        self.update_plot_data()

    def update_plot_data(self):
        try:
            self.plot.update_plot_data(self.stats_data, stats_data_2=self.stats_data_2, included_vars=self.selections)
        except Exception as e:
            msg = "Correlation calculation failed. Perhaps more data is needed?\n%s" % str(e)
            ErrorDialog(self.parent, msg, "Correlation Error")

    def clear_data(self):
        pass

    def get_csv(self, selection=None):
        pass

    def export_csv(self, evt):
        save_data_to_file(self.parent, "Export Correlation data to CSV", self.plot.get_csv())

    def save_html(self, *evt):
        save_data_to_file(self.parent, 'Save Correlation to HTML', self.plot.html_str, initial_dir="",
                          wildcard="HTML files (*.html)|*.html")

    def save_svg(self, *evt):
        try:
            save_data_to_file(self.parent, 'Save Correlation to SVG', self.plot.export_svg,
                              initial_dir="", data_type='function', wildcard="SVG files (*.svg)|*.svg")
        except Exception as e:
            ErrorDialog(self.parent, str(e), "Save Error")





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.roi_name_manager.py
"""
Code to create and edit a the roi name mapping, calculate data used to plot map

"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import os
from copy import deepcopy
import difflib
from dvha.db.sql_to_python import QuerySQL
from dvha.db.sql_connector import DVH_SQL
from dvha.paths import PREF_DIR
from dvha.tools.roi_map_generator import ROIMapGenerator
from dvha.tools.utilities import flatten_list_of_lists, initialize_directories


class PhysicianROI:
    def __init__(self, physician_roi, institutional_roi=None):
        self.institutional_roi = institutional_roi if institutional_roi is not None else 'uncategorized'
        self.physician_roi = physician_roi
        self.variations = [physician_roi]

        self.add_variations(self.institutional_roi)

    def __contains__(self, variation):
        return clean_name(variation) in self.clean_variations

    def add_variations(self, variations):
        if type(variations) is not list:
            variations = [variations]
        clean_variations = self.clean_variations
        for variation in variations:
            if clean_name(variation) not in clean_variations and variation.lower() not in {'uncategorized'}:
                self.variations.append(variation)

    def get_variation(self, variation):
        clean_variation = variation
        clean_variations = self.clean_variations
        if clean_variation in clean_variations:
            index = clean_variations.index(clean_variation)
            return self.variations[index]

    @property
    def clean_variations(self):
        return [clean_name(v) for v in self.variations]

    @property
    def clean_top_level(self):
        return [clean_name(v) for v in [self.institutional_roi, self.physician_roi]]

    def del_variation(self, variations):
        if type(variations) is not list:
            variations = list(variations)
        clean_variations = self.clean_variations
        for variation in variations:
            clean_variation = clean_name(variation)
            if clean_variation in clean_variations and clean_variation not in self.clean_top_level:
                index = clean_variations.index(clean_variation)
                self.variations.pop(index)

    def del_all_variations(self):
        self.variations = [self.physician_roi]
        if self.institutional_roi.lower() != 'uncategorized' and \
                clean_name(self.institutional_roi) != clean_name(self.physician_roi):
            self.variations.append(self.institutional_roi)

    def set_institutional_roi(self, new):
        clean_old = clean_name(self.institutional_roi)
        if clean_old in self.clean_variations and \
                clean_old != clean_name(self.physician_roi):
            index = self.clean_variations.index(clean_old)
            self.variations.pop(index)

        self.institutional_roi = new
        self.add_variations(new)


class Physician:
    """Represents a physician in the roi map"""
    def __init__(self, name):
        """
        :param name: the label to be used to represent the physician, should be all upper case
        :type name: str
        """
        self.name = name
        self.rois = {}

    def __contains__(self, variation):
        return clean_name(variation) in self.all_clean_variations

    def add_physician_roi(self, institutional_roi, physician_roi, variations=None):
        self.rois[physician_roi] = PhysicianROI(physician_roi, institutional_roi)
        if variations is not None:
            self.add_variations(physician_roi, variations)

    def del_physician_roi(self, physician_roi):
        clean_roi = clean_name(physician_roi)
        clean_rois_map = self.clean_physician_rois_map
        if clean_roi in list(clean_rois_map):
            self.rois.pop(clean_rois_map[clean_roi], None)

    def add_variations(self, physician_roi, variations):
        if physician_roi in list(self.rois):
            self.rois[physician_roi].add_variations(variations)

    def del_variations(self, physician_roi, variations):
        if physician_roi in list(self.rois):
            self.rois[physician_roi].del_variation(variations)

    def delete_all_variations(self, physician_roi=None):
        physician_rois = [physician_roi] if physician_roi is None else list(self.rois)
        for physician_roi in physician_rois:
            self.rois[physician_roi].del_all_variations()

    @property
    def all_variations(self):
        variations = []
        for physician_roi in self.rois.values():
            variations.extend(physician_roi.variations)
        return variations

    def get_variations(self, physician_roi):
        clean_physician_roi = clean_name(physician_roi)
        clean_map = self.clean_physician_rois_map
        if clean_physician_roi in list(clean_map):
            return self.rois[clean_map[clean_physician_roi]].variations

    @property
    def all_clean_variations(self):
        return [clean_name(v) for v in self.all_variations]

    @property
    def institutional_rois(self):
        return [physician_roi.institutional_roi for physician_roi in self.rois.values()]

    @property
    def clean_institutional_rois_map(self):
        return {clean_name(roi): roi for roi in self.institutional_rois}

    @property
    def clean_physician_rois(self):
        return [clean_name(roi) for roi in list(self.rois)]

    @property
    def clean_physician_rois_map(self):
        return {clean_name(roi): roi for roi in self.rois}

    def get_institutional_roi(self, variation, return_physician_roi=False):
        for physician_roi_str, physician_roi in self.rois.items():
            if variation in physician_roi:
                if return_physician_roi:
                    return physician_roi.institutional_roi, physician_roi_str
                return physician_roi.institutional_roi

        if return_physician_roi:
            return None, None

    def rename_institutional_roi(self, new, old):
        _, physician_roi_str = self.get_institutional_roi(old, return_physician_roi=True)
        if physician_roi_str is not None:
            self.rois[physician_roi_str].set_institutional_roi(new)

    def get_physician_roi(self, variation):
        for physician_roi_str, physician_roi in self.rois.items():
            if variation in physician_roi:
                return physician_roi_str
        return 'uncategorized'

    def is_physician_roi(self, physician_roi):
        return clean_name(physician_roi) in self.clean_physician_rois

    def rename_physician_roi(self, new, old):
        clean_old = clean_name(old)
        clean_rois = self.clean_physician_rois_map
        if clean_old in list(clean_rois):
            actual_old = clean_rois[clean_old]
            self.rois[new] = self.rois.pop(actual_old)
            self.rois[new].add_variations(new)
            self.rois[new].physician_roi = new

    def get_variation(self, variation):
        for physician_roi in self.rois.values():
            ans = physician_roi.get_variation(variation)
            if ans is not None:
                return ans


class DatabaseROIs:
    """The main class, creating an instance of this class will open a stored map, or create the default map"""
    def __init__(self):

        self.physicians = {}
        self.institutional_rois = []

        if not os.path.isdir(PREF_DIR):
            initialize_directories()

        self.physicians_from_file = get_physicians_from_roi_files()

        self.branched_institutional_rois = {}
        self.import_from_file()

    def import_from_file(self):
        self.physicians = {}
        self.institutional_rois = []

        if 'DEFAULT' in self.physicians_from_file:
            abs_file_path = os.path.join(PREF_DIR, 'physician_DEFAULT.roi')
            self.import_physician_roi_map(abs_file_path, 'DEFAULT')
        else:
            self.institutional_rois = ROIMapGenerator().primary_names
            self.add_physician('DEFAULT')

        for physician in self.physicians_from_file:
            self.add_physician(physician)

        self.import_physician_roi_maps()

        if 'uncategorized' not in self.institutional_rois:
            self.institutional_rois.append('uncategorized')

        self.branched_institutional_rois = {}

    ##############################################
    # Import from file functions
    ##############################################
    def import_physician_roi_maps(self):

        for physician in get_physicians_from_roi_files():
            rel_path = 'physician_%s.roi' % physician
            abs_file_path = os.path.join(PREF_DIR, rel_path)
            if os.path.isfile(abs_file_path):
                self.import_physician_roi_map(abs_file_path, physician)

    def import_physician_roi_map(self, abs_file_path, physician=None):

        institutional_mode = physician == 'DEFAULT'

        if physician is None:
            physician = os.path.splitext(os.path.basename(abs_file_path))[0].split('physician_')[1]

        if physician not in list(self.physicians):
            self.add_physician(physician)

        with open(abs_file_path, 'r') as document:
            for line in document:
                if not line:
                    continue
                line = str(line).strip().replace(':', ',').split(',')
                institutional_roi = line.pop(0)
                if institutional_mode:
                    self.add_institutional_roi(institutional_roi)
                else:
                    physician_roi = line.pop(0)
                    variations = [v.strip() for v in line]

                    self.add_physician_roi(physician, institutional_roi.strip(), physician_roi.strip(), variations)

    ###################################
    # Physician functions
    ###################################
    def get_physicians(self):
        return ['DEFAULT'] + sorted(set(self.physicians) - {'DEFAULT'})

    def add_physician(self, physician):
        physician = clean_name(physician, physician=True)
        if physician not in list(self.physicians):
            self.physicians[physician] = Physician(physician)

        if physician == 'DEFAULT':
            for institutional_roi in self.institutional_rois:
                self.add_physician_roi(physician, institutional_roi, institutional_roi)

    def copy_physician(self, new_physician, copy_from=None, include_variations=True):
        new_physician = clean_name(new_physician, physician=True)
        if copy_from is None or copy_from == 'DEFAULT':
            self.add_physician(new_physician)
        elif copy_from in list(self.physicians):
            self.physicians[new_physician] = deepcopy(self.physicians[copy_from])
            if not include_variations:
                self.physicians[new_physician].delete_all_variations()

    def delete_physician(self, physician):
        if physician in list(self.physicians):
            self.physicians.pop(physician, None)

    def is_physician(self, physician):
        return physician in list(self.physicians)

    def rename_physician(self, new_physician, physician):
        new_physician = clean_name(new_physician, physician=True)
        physician = clean_name(physician, physician=True)
        self.physicians[new_physician] = self.physicians.pop(physician)

    #################################
    # Institutional ROI functions
    #################################
    @property
    def clean_institutional_rois(self):
        return [clean_name(roi) for roi in self.institutional_rois]

    def get_institutional_roi(self, physician, physician_roi):
        if physician in list(self.physicians):
            if physician_roi in self.physicians[physician]:
                return self.physicians[physician].get_institutional_roi(physician_roi)

    def add_institutional_roi(self, roi):
        if roi not in self.institutional_rois:
            self.institutional_rois.append(roi)
            self.add_physician_roi('DEFAULT', roi, roi)

    def rename_institutional_roi(self, new, old):
        if old in self.institutional_rois:
            index = self.institutional_rois.index(old)
            self.institutional_rois.pop(index)
            for physician in self.physicians.values():
                physician.rename_institutional_roi(new, old)
            if new != 'uncategorized':
                self.add_institutional_roi(new)

    def set_linked_institutional_roi(self, new_institutional_roi, physician, physician_roi):
        if physician in list(self.physicians) and physician_roi in list(self.physicians[physician].rois):
            self.physicians[physician].rois[physician_roi].set_institutional_roi(new_institutional_roi)

    def delete_institutional_roi(self, roi):
        self.rename_institutional_roi('uncategorized', roi)
        self.delete_physician_roi('DEFAULT', roi)

    def is_institutional_roi(self, roi):
        return clean_name(roi) in self.clean_institutional_rois

    def get_unused_institutional_rois(self, physician):
        physician = clean_name(physician, physician=True)
        used_rois = []
        if physician in list(self.physicians):
            used_rois = self.physicians[physician].institutional_rois

        unused_rois = []
        for roi in self.institutional_rois:
            if roi not in used_rois:
                unused_rois.append(roi)
        if 'uncategorized' not in unused_rois:
            unused_rois.append('uncategorized')

        return unused_rois

    ########################################
    # Physician ROI functions
    ########################################
    def get_physician_rois(self, physician):
        physician = clean_name(physician, physician=True)
        if physician in list(self.physicians):
            return sorted(list(self.physicians[physician].rois))
        return []

    def get_physician_roi(self, physician, roi):
        physician = clean_name(physician, physician=True)
        if physician in self.physicians:
            ans = self.physicians[physician].get_physician_roi(roi)
            if ans:
                return ans
        return 'uncategorized'

    def get_physician_roi_from_institutional_roi(self, physician, institutional_roi):
        physician = clean_name(physician, physician=True)
        if physician in list(self.physicians):
            return self.physicians[physician].get_physician_roi(institutional_roi)

    def add_physician_roi(self, physician, institutional_roi, physician_roi, variations=None):
        self.add_institutional_roi(institutional_roi)  # will skip if already exists
        if physician in list(self.physicians):
            self.physicians[physician].add_physician_roi(institutional_roi, physician_roi, variations)

    def rename_physician_roi(self, new_physician_roi, physician, physician_roi):
        if physician in list(self.physicians):
            self.physicians[physician].rename_physician_roi(new_physician_roi, physician_roi)

    def delete_physician_roi(self, physician, physician_roi):
        physician = clean_name(physician, physician=True)
        if physician in list(self.physicians):
            self.physicians[physician].del_physician_roi(physician_roi)

    def is_physician_roi(self, roi, physician):
        physician = clean_name(physician, physician=True)
        if physician in list(self.physicians):
            return self.physicians[physician].is_physician_roi(roi)

    def get_unused_physician_rois(self, physician):
        physician = clean_name(physician, physician=True)

        unused_rois = []
        for physician_roi in self.get_physician_rois(physician):
            if self.get_institutional_roi(physician, physician_roi) == 'uncategorized':
                unused_rois.append(physician_roi)

        return unused_rois

    def merge_physician_rois(self, physician, physician_rois, final_physician_roi):

        variation_lists = [self.get_variations(physician, physician_roi) for physician_roi in physician_rois]
        variations = flatten_list_of_lists(variation_lists, remove_duplicates=True)
        self.add_variations(physician, final_physician_roi, variations)

        for physician_roi in physician_rois:
            if physician_roi != final_physician_roi:
                self.delete_physician_roi(physician, physician_roi)

    ###################################################
    # Variation-of-Physician-ROI functions
    ###################################################
    def get_variations(self, physician, physician_roi):
        physician = clean_name(physician, physician=True)
        physician_roi = clean_name(physician_roi)
        if physician_roi == 'uncategorized':
            return ['uncategorized']

        if physician in list(self.physicians):
            variations = self.physicians[physician].get_variations(physician_roi)
            if variations:
                return variations
        return []

    def get_all_variations_of_physician(self, physician):
        physician = clean_name(physician, physician=True)
        if physician in list(self.physicians):
            variations = self.physicians[physician].all_variations
            if variations:
                return variations
        return []

    def is_variation_used(self, physician, variation):
        if physician in list(self.physicians):
            return variation in self.physicians[physician]
        return False

    def add_variations(self, physician, physician_roi, variation):
        physician = clean_name(physician, physician=True)
        if physician in list(self.physicians):
            self.physicians[physician].add_variations(physician_roi, variation)

    def delete_variations(self, physician, physician_roi, variations):
        physician = clean_name(physician, physician=True)
        if physician in list(self.physicians):
            if type(variations) is not list:
                variations = [variations]
            self.physicians[physician].del_variations(physician_roi, variations)

    def set_variation(self, new_variation, physician, physician_roi, variation):
        physician = clean_name(physician, physician=True)
        if new_variation != variation:
            self.add_variations(physician, physician_roi, new_variation)
            self.delete_variations(physician, physician_roi, variation)

    def is_roi(self, roi):
        if clean_name(roi) in self.clean_institutional_rois:
            return True
        for physician in self.physicians.values():
            if roi in physician:
                return True
        return False

    ########################
    # Export to file
    ########################
    def write_to_file(self):
        for physician, data in self.physician_roi_file_data.items():
            self.write_physician_file(physician, data)
        self.remove_unused_roi_files()

    @property
    def physician_roi_file_data(self):
        physicians_file_data = {}
        for physician in list(self.physicians) + ['DEFAULT']:
            lines = []
            for physician_roi in self.get_physician_rois(physician):
                institutional_roi = self.get_institutional_roi(physician, physician_roi)
                variations = ', '.join(self.get_variations(physician, physician_roi))
                lines.append(': '.join([institutional_roi, physician_roi, variations]))
            lines.sort()
            physicians_file_data[physician] = lines

        return physicians_file_data

    @staticmethod
    def write_physician_file(physician, lines):
        """
        Write the physician map to a .roi file
        :param physician: name of physicain
        :type physician: str
        :param lines: the lines of data to be written to the file
        :type lines: list of str
        """
        abs_file_path = os.path.join(PREF_DIR, 'physician_' + physician + '.roi')
        if lines:
            with open(abs_file_path, 'w') as document:
                for line in lines:
                    document.write(line + '\n')

    def remove_unused_roi_files(self):
        """
        Delete any physician .roi files that are no longer in the ROI map
        :return: the physicians that have been removed
        :rtype: list
        """
        for physician in self.deleted_physicians:
            file_name = 'physician_' + physician + '.roi'
            abs_file_path = os.path.join(PREF_DIR, file_name)
            os.remove(abs_file_path)

    @property
    def deleted_physicians(self):
        return list(set(get_physicians_from_roi_files()) - set(self.get_physicians()))

    @property
    def added_physicians(self):
        return list(set(self.get_physicians()) - set(get_physicians_from_roi_files()))

    def get_roi_map_changes(self):
        """
        Use difflib to detect changes between current roi map and previously saved .roi file
        format of returned dict: diff[physician][physician_roi][delta] = {'institutional': i_roi, 'variations': list}
        where delta is either '+' or '-' based on difflib.unified_diff output
        :return: a tiered dictionary of lines that changed in the proposed .roi file
        :rtype: dict
        """
        new_data = self.physician_roi_file_data
        diff = {}
        for physician in list(self.physicians):
            abs_file_path = os.path.join(PREF_DIR, 'physician_' + physician + '.roi')
            if os.path.isfile(abs_file_path):
                old_lines = [line.strip() for line in open(abs_file_path, 'r').readlines()]

                include = False
                diff[physician] = {}
                for line in difflib.unified_diff(old_lines, new_data[physician]):
                    if include:
                        if line[0] in {'+', '-'}:
                            i_roi, p_roi, variations = tuple(i for i in line.split(': '))
                            if p_roi not in diff[physician]:
                                diff[physician][p_roi] = {'-': {'institutional': '', 'variations': []},
                                                          '+': {'institutional': '', 'variations': []}}
                            diff[physician][p_roi][line[0]] = {'institutional': i_roi[1:],
                                                               'variations': variations.split(', ')}
                    else:
                        include = line[0] == '@'  # + and - signs before the @@ line to be ignored

                for p_roi, data in diff[physician].items():
                    new_pos = list(set(data['+']['variations']) - set(data['-']['variations']))
                    new_neg = list(set(data['-']['variations']) - set(data['+']['variations']))
                    if new_pos:
                        data['+']['variations'] = new_pos
                    else:
                        data.pop('+')
                    if new_neg:
                        data['-']['variations'] = new_neg
                    else:
                        data.pop('-')

        return diff

    @property
    def physicians_to_remap(self):
        return list(set(self.deleted_physicians + self.added_physicians))

    @property
    def variations_to_update(self):
        """
        :return: all variations that have changed physician or institutional rois, in a dict with physician for the key
        :rtype: dict
        """
        changes = self.get_roi_map_changes()
        variations_to_update = {}
        for physician, physician_roi_data in changes.items():
            variations = []
            for p_roi_data in physician_roi_data.values():
                for delta_data in p_roi_data.values():
                    variations.extend(delta_data['variations'])
            variations_to_update[physician] = list(set(variations))

        for physician in list(variations_to_update):
            if not variations_to_update[physician]:
                variations_to_update.pop(physician)

        return variations_to_update

    def remap_rois(self):

        with DVH_SQL() as cnx:
            for physician, variations in self.variations_to_update.items():
                for variation in variations:
                    new_physician_roi = self.get_physician_roi(physician, variation)
                    new_institutional_roi = self.get_institutional_roi(physician, new_physician_roi)

                    condition = "REPLACE(REPLACE(LOWER(roi_name), '\'', '`'), '_', ' ') == '%s'" % variation
                    sql_query = "SELECT DISTINCT study_instance_uid, roi_name FROM DVHs WHERE %s;" % condition
                    uids_roi_names = cnx.query_generic(sql_query)
                    uids = [row[0] for row in uids_roi_names]
                    roi_names = [row[1] for row in uids_roi_names]

                    if uids:
                        for i, uid in enumerate(uids):
                            roi_name = roi_names[i]
                            condition = "roi_name = '%s' and study_instance_uid = '%s'" % (roi_name, uid)
                            cnx.update('dvhs', 'physician_roi', new_physician_roi, condition)
                            cnx.update('dvhs', 'institutional_roi', new_institutional_roi, condition)

        self.write_to_file()

    ################
    # Plotting tools
    ################
    def get_physician_roi_visual_coordinates(self, physician, physician_roi, institutional_roi):

        # All 0.5 subtractions due to a workaround of a Bokeh 0.12.9 bug

        # x and y are coordinates for the circles
        # x0, y0 is beginning of line segment, x1, y1 is end of line-segment
        if institutional_roi == 'uncategorized':
            table = {'name': [physician_roi],
                     'x': [1.5],
                     'y': [0],
                     'x0': [1.5],
                     'y0': [0],
                     'x1': [1.5],
                     'y1': [0]}
        else:
            table = {'name': [institutional_roi, physician_roi],
                     'x': [0.5, 1.5],
                     'y': [0, 0],
                     'x0': [0.5, 1.5],
                     'y0': [0, 0],
                     'x1': [1.5, 0.5],
                     'y1': [0, 0]}

        variations = self.physicians[physician].rois[physician_roi].variations
        for i, variation in enumerate(variations):
            table['name'].append(variation)
            table['x'].append(2.5)
            table['y'].append(-i)
            table['x0'].append(1.5)
            table['y0'].append(0)
            table['x1'].append(2.5)
            table['y1'].append(-i)

        table_length = len(table['name'])
        table['color'] = ['#1F77B4'] * table_length
        table['institutional_roi'] = [institutional_roi] * table_length
        table['physician_roi'] = [physician_roi] * table_length

        return table

    def get_all_institutional_roi_visual_coordinates(self, physician, ignored_physician_rois=None):
        # TODO: Although functional and faster, hard to follow
        # Still slowest part of ROI Map plot generation
        ignored_physician_rois = [] if ignored_physician_rois is None else ignored_physician_rois
        p_and_i = [(roi.physician_roi, roi.institutional_roi)
                   for roi in self.physicians[physician].rois.values()
                   if roi.physician_roi not in ignored_physician_rois]

        i_rois = [roi[1] for roi in p_and_i]
        i_rois.sort()
        for i, i_roi in enumerate(i_rois):
            if i_roi == 'uncategorized':
                i_rois[i] = 'zzzzzzzzzzzzzzzzzzz'
        sorted_indices = [i[0] for i in sorted(enumerate(i_rois), key=lambda x:x[1])]
        p_and_i = [p_and_i[i] for i in sorted_indices]

        tables = {roi[0]: self.get_physician_roi_visual_coordinates(physician, roi[0], roi[1])
                  for roi in p_and_i}
        heights = [3 - min(tables[p_roi[0]]['y']) for p_roi in p_and_i]
        max_y_delta = sum(heights) + 2  # include 2 buffer to give space to read labels on plot
        for i, roi in enumerate(p_and_i):
            y_delta = sum(heights[i:])

            for key in ['y', 'y0', 'y1']:
                for j in range(len(tables[roi[0]][key])):
                    tables[roi[0]][key][j] += y_delta - max_y_delta

        i_roi_map = {p_i[0]: p_i[1] for p_i in p_and_i}
        if p_and_i and p_and_i[0][0] in tables:
            table = tables[p_and_i[0][0]]
            for i in range(1, len(p_and_i)):
                for key in list(table):
                    table[key].extend(tables[p_and_i[i][0]][key])

            return self.update_duplicate_y_entries(table, physician, i_roi_map)
        return None

    @staticmethod
    def get_roi_visual_y_values(table):
        y_values = {}
        for i, x in enumerate(table['x']):
            if x == 0.5:
                name = table['name'][i]
                if name not in list(y_values):
                    y_values[name] = []
                y_values[name].append(table['y'][i])
        for name in list(y_values):
            y_values[name] = sum(y_values[name]) / len(y_values[name])
        return y_values

    def update_duplicate_y_entries(self, table, physician, inst_map):

        y_values = self.get_roi_visual_y_values(table)

        self.branched_institutional_rois[physician] = []

        for i, name in enumerate(table['name']):
            if table['x'][i] == 0.5 and table['y'][i] != y_values[name]:
                table['y'][i] = y_values[name]
                table['y0'][i] = y_values[name]
                table['color'][i] = 'red'
                self.branched_institutional_rois[physician].append(name)
            if table['x'][i] == 1.5:
                inst_name = inst_map[name]
                if inst_name != 'uncategorized':
                    table['y1'][i] = y_values[inst_name]

        if self.branched_institutional_rois[physician]:
            self.branched_institutional_rois[physician] = list(set(self.branched_institutional_rois[physician]))

        return table

    @property
    def tree(self):
        return {physician: self.get_physician_tree(physician) for physician in self.get_physicians()}

    def get_physician_tree(self, physician):
        phys_rois = self.get_physician_rois(physician)
        unused_inst_rois = self.get_unused_institutional_rois(physician)
        inst_rois = [roi for roi in self.institutional_rois if roi not in unused_inst_rois]
        linked_phys_rois = [self.get_physician_roi_from_institutional_roi(physician, roi) for roi in inst_rois]
        unlinked_phys_rois = [roi for roi in phys_rois if roi not in linked_phys_rois]
        linked_phys_roi_tree = {roi: self.get_variations(physician, roi) for roi in linked_phys_rois
                                if roi != 'uncategorized'}
        unlinked_phys_roi_tree = {roi: self.get_variations(physician, roi) for roi in unlinked_phys_rois
                                  if roi != 'uncategorized'}
        return {'Linked to Institutional ROI': linked_phys_roi_tree,
                'Unlinked to Institutional ROI': unlinked_phys_roi_tree}


def clean_name(name, physician=False):
    ans = str(name).replace('\'', '`').replace('_', ' ').strip()
    while '  ' in ans:
        ans = ans.replace('  ', ' ')

    if physician:
        return ans.replace(' ', '_').upper()
    return ans.lower()


def get_physicians_from_roi_files():

    physicians = []
    for file_name in os.listdir(PREF_DIR):
        if file_name.startswith("physician_") and file_name.endswith(".roi"):
            physician = file_name.replace('physician_', '').replace('.roi', '')
            physician = clean_name(physician, physician=True)
            physicians.append(physician)

    return physicians


def get_physician_from_uid(uid):
    with DVH_SQL() as cnx:
        results = cnx.query('Plans', 'physician', "study_instance_uid = '" + uid + "'")

    if len(results) > 1:
        print('Warning: multiple plans with this study_instance_uid exist')

    return str(results[0][0])


def update_uncategorized_rois_in_database():
    roi_map = DatabaseROIs()
    dvh_data = QuerySQL('DVHs', "physician_roi = 'uncategorized'")

    with DVH_SQL() as cnx:
        for i in range(len(dvh_data.roi_name)):
            uid = dvh_data.study_instance_uid[i]
            mrn = dvh_data.mrn[i]
            physician = get_physician_from_uid(uid)
            roi_name = dvh_data.roi_name[i]

            new_physician_roi = roi_map.get_physician_roi(physician, roi_name)
            new_institutional_roi = roi_map.get_institutional_roi(physician, roi_name)

            if new_physician_roi != 'uncategorized':
                print(mrn, physician, new_institutional_roi, new_physician_roi, roi_name, sep=' ')
                condition = "study_instance_uid = '" + uid + "'" + "and roi_name = '" + roi_name + "'"
                cnx.update('DVHs', 'physician_roi', new_physician_roi, condition)
                cnx.update('DVHs', 'institutional_roi', new_institutional_roi, condition)
